# coding=utf-8

import sys, os, re
from argparse import ArgumentParser, Namespace, _ArgumentGroup, _SubParsersAction
from LRphase.SimulatePhasedData import *
from LRphase.InputData import *
from LRphase import urls
from pysam import AlignmentFile, VariantFile
import numpy as np
#from plfit import plfit
import powerlaw
import time
from LRphase.PhaseSet import powlaw_modified, _estimate_prior_probabilities

__version__ = "0.7.11"

def phasing(args):
    """Main function for phasing mode

    Args:
        args:
    """
    sys.stderr.write('\nAnalyzing inputs using phasing mode...\n\n')
    #sys.stderr.write("%s\n" % args)
    # Input data initialization currently ignores much of the content of args. May be a good idea to fix this.
    # TO-DO: Need to actually handle VCF samples properly instead of relying on args.one_sample.
    input_data = InputData(args.output_directory_name,
                           sample = args.one_sample,
                           ignore_samples = args.ignore_samples,
                           reference_sequence_input = args.reference_genome,
                           reference_sequence_input_assembly = args.reference_assembly,
                           max_threads = args.threads,
                           quiet = args.quiet_mode,
                           silent = args.silent_mode
    )
    input_data.add_haplotype_information(args.vcf_file_name)
    input_data.add_reference_sequence(args.reference_genome,
                                      reference_sequence_input_assembly = args.reference_assembly,
                                      output_directory = args.output_directory_name)
    try:
        input_data.add_reads(args.long_read_inputs, sample = args.one_sample, quiet = args.quiet_mode, silent = args.silent_mode)
    except OSError as e:
        sys.stderr.write("%s\n" % e())
        exit(2)
    sys.stderr.write("Phasing reads...\n")
    i=0
    for phasable_sample in input_data:
        sys.stderr.write('Processing alignments for sample %d (%s)...\n' % (i, phasable_sample.sample))
        header_template_file = AlignmentFile(phasable_sample.alignment_file_paths[i], 'rb')

        # Instantiate output file(s)
        if args.output_mode == 'combined':
            combined_output_bam = AlignmentFile('%s/%s.combined_phasing_results.bam' % (args.output_directory_name, phasable_sample.sample),
                                                'wb', template = header_template_file)
        else:
            unphased_output_bam = AlignmentFile('%s/%s.unphased.bam' %  (args.output_directory_name, phasable_sample.sample),
                                                'wb', template = header_template_file)
            nonphasable_output_bam = AlignmentFile('%s/%s.nonphasable.bam' %  (args.output_directory_name, phasable_sample.sample),
                                                   'wb', template = header_template_file)
            if args.output_mode == 'phase_tagged':
                phased_output_bam = AlignmentFile('%s/%s.phase_tagged.bam' %  (args.output_directory_name, phasable_sample.sample),
                                                  'wb', template = header_template_file)
            elif args.output_mode == 'full':
                maternal_output_bam = AlignmentFile('%s/%s.maternal.bam' %  (args.output_directory_name, phasable_sample.sample),
                                                    'wb', template = header_template_file)
                paternal_output_bam = AlignmentFile('%s/%s.paternal.bam' %  (args.output_directory_name, phasable_sample.sample),
                                                    'wb', template = header_template_file)
                
        # Phase reads and write output to approriate file(s)
        for alignment in phasable_sample:
            phased_read = PhasedRead(alignment,
                                     vcf_file = phasable_sample.vcf_file_path,
                                     sample = phasable_sample.sample,
                                     powlaw_alpha = args.powlaw_alpha,
                                     powlaw_xmin = args.powlaw_xmin
            )

            # Make sure reads are tagged appropriately, according to output options.
            if args.add_phasing_tag_to_output:
                # Tag with estimated error rate based on fitted power-law model.
                phased_read.aligned_segment.set_tag(tag = 'ER', value = str(phased_read.phasing_error_rate), value_type='Z', replace=True)
                if phased_read.is_phased:
                    if (args.log_likelihood_threshold >= 0 and phased_read.log_likelihood_ratio >= args.log_likelihood_threshold) or phased_read.phasing_error_rate <= args.error_rate_threshold:
                        # Phased read passing the error rate/score threshold.
                        pass
                    else:
                        # Phasing score does not pass threshold. Label as unphased.
                        phased_read.aligned_segment.set_tag(tag = 'HP', value = "unphased", value_type='Z', replace=True)
                elif phased_read.is_Unphased:
                    phased_read.aligned_segment.set_tag(tag = 'HP', value = "unphased", value_type='Z', replace=True)
                else: # Nonphasable
                    phased_read.aligned_segment.set_tag(tag = 'HP', value = "nonphasable", value_type='Z', replace=True)
                    
            # Write output to appropriate destination file(s)
            if args.output_mode == 'combined':
                write_combined_bam_output(phased_read, combined_output_bam, args)
            elif args.output_mode == 'phase_tagged':
                write_phase_tagged_bam_output(phased_read, phased_output_bam, unphased_output_bam,
                                              nonphasable_output_bam, args)
            elif args.output_mode == 'full':
                write_full_bam_output(phased_read, maternal_output_bam, paternal_output_bam,
                                      unphased_output_bam, nonphasable_output_bam, args)                
            
        # Close output file(s)(
        for outfile in ["combined_output_bam", "unphased_output_bam", "nonphasable_output_bam",
                        "phased_output_bam", "maternal_output_bam", "paternal_output_bam"]:
            if outfile in locals():
                locals()[outfile].close()
                
        i += 1
        
    return


def write_combined_bam_output(phased_read, combined_output_bam, args):
    """
    Write phased, unphased, and nonphasable reads to a single output file.
    """
    phased_read.write_to_bam(output_bam_pysam = combined_output_bam)
    return


def write_phase_tagged_bam_output(phased_read, phased_output_bam,
                                  unphased_output_bam, nonphasable_output_bam, args):
    """
    Write phased reads to a single output file, with unphased and nonphasable
    reads written to their own files.

    """
    if phased_read.is_phased:
        if (args.log_likelihood_threshold >= 0 and phased_read.log_likelihood_ratio >= args.log_likelihood_threshold) or phased_read.phasing_error_rate <= args.error_rate_threshold:
            # Phased read passing the error-rate/score threshold.
            phased_read.write_to_bam(output_bam_pysam = phased_output_bam)
        else:
            # Phasing score does not pass threshold. Treat as unphased.
            phased_read.write_to_bam(output_bam_pysam = unphased_output_bam)
    elif phased_read.is_Unphased:
        phased_read.write_to_bam(output_bam_pysam = unphased_output_bam)
    else: # Nonphasable
        phased_read.write_to_bam(output_bam_pysam = nonphasable_output_bam)
    return
            

def write_full_bam_output(phased_read, maternal_output_bam, paternal_output_bam,
                          unphased_output_bam, nonphasable_output_bam, args):
    """
    Write maternal and paternal phased reads to separate output files, with 
    unphased and nonphasable reads written to their own files.
    """
    # TO-DO: Need to handle all possible levels of ploidy.
    if phased_read.is_phased:
        if (args.log_likelihood_threshold >= 0 and phased_read.log_likelihood_ratio >= args.log_likelihood_threshold) or phased_read.phasing_error_rate <= args.error_rate_threshold:
            if phased_read.phase == 1:
                phased_read.write_to_bam(output_bam_pysam = paternal_output_bam)
            else:
                phased_read.write_to_bam(output_bam_pysam = maternal_output_bam)
        else:
            # Phasing score does not pass threshold. Label as unphased and handle accordingly.
            phased_read.write_to_bam(output_bam_pysam = unphased_output_bam)
    elif phased_read.is_Unphased:
        phased_read.write_to_bam(output_bam_pysam = unphased_output_bam)
    else: # Nonphasable
        phased_read.write_to_bam(output_bam_pysam = nonphasable_output_bam)
    return



def phasability(args):
    """Main function for phasability mode

    Args:
        args:
    """
    
    print('phasability mode')
    #print(args)

    # First precompute error rates over a (un)reasonable range of heterozygous site counts.
    max_hets = 1000
    prior_probs = _estimate_prior_probabilities()
    _error_rate_table = _calculate_error_rate_table(max_hets, args.sequencing_error_rate,
                                                    prior_probs, args.powlaw_alpha, args.powlaw_xmin)

    # Load up the variants
    vcf_file = VariantFile(args.vcf_file_name)

    # Load up chrom_sizes/intervals
    if args.regions:
        intervals = _build_intervals_from_bed_file(args.regions)
    else:
        intervals = _build_intervals_from_chrom_sizes_dict(_chrom_sizes_dict_from_file(args.chrom_sizes))

    # Step through windows, pulling out variant counts and
    # returning the associated error rates.
    with open("%s/estimated_phasability.wig" % args.output_directory_name, 'w') as outfile:
        # Print the wiggle file header
        outfile.write('track type=wig name="Phasability" description="Phasability Plot"\n')
        for interval in intervals:
            chrom, int_start, int_end = interval

            # Make sure chromosome is in VCF
            if chrom not in list(vcf_file.header.contigs):
                continue
            
            # Print wig header
            outfile.write("fixedStep chrom=%s start=%s step=%d span=%d\n" % (chrom, int_start, args.step_size, args.step_size))

            # Step through windows within the chromosome
            win_start = int_start
            win_end = win_start + args.window_size
            while win_start < win_end and win_end <= int_end:
                n_hets = len([het for het in vcf_file.fetch(chrom, win_start, win_end)])
                if n_hets > max_hets:
                    # Have to calculate if more hets than anticipated
                    error_rate = _calculate_error_rate_from_het_count(n_hets, args.sequencing_error_rate,
                                                                      prior_probs,
                                                                      args.powlaw_alpha, args.powlaw_xmin)
                else:
                    # Use precomputed values
                    error_rate = _error_rate_table[n_hets]
                outfile.write("%.5f\n" % (1-error_rate))

                win_start += args.step_size
                win_end = win_start + args.window_size
                if win_end > int_end:
                    win_end = int_end
            
    # Clean up after ourselves
    vcf_file.close()    
    
    return


def _build_intervals_from_bed_file(bed_file):
    """
    Build a list of intervals (simply tuples of chrom, start, end) from
    a BED file.
    """
    ret = []
    with open(bed_file, 'r') as infile:
        for line in infile:
            if re.match('^#', line):  # Comment line
                continue
            rec = line.strip('\n').split()
            ret.append((rec[0], rec[1], rec[2]))
    return ret


def _build_intervals_from_chrom_sizes_dict(chrom_sizes):
    """
    Build a list of intervals from the chrom_sizes dict.
    """
    ret = []
    for chrom in chrom_sizes.keys():
        ret.append((chrom, 1, chrom_sizes[chrom]))
    return ret


def _chrom_sizes_dict_from_file(chrom_sizes):
    """
    Load chrom_sizes entries from a file into a dictionary.
    """
    chrom_sizes_dict = {}
    with open(chrom_sizes, 'r') as chrom_sizes_file:
        for line in chrom_sizes_file:
            rec = line.strip('\n').split()
            chrom_sizes_dict[rec[0]] = int(rec[1])
    return(chrom_sizes_dict)
    

def _calculate_error_rate_from_het_count(het_count, sequencing_error_rate, prior_probs, powlaw_alpha, powlaw_xmin):
    """
    Caclulate an empirical error rate given a discrete het count, sequencing
    error rate, bayesian priors, and powlaw alpha and xmin.
    """
    if het_count <= powlaw_xmin:
        return 1
    x = (het_count  * np.log10(1 - sequencing_error_rate)) + max(prior_probs)
    y = (het_count * np.log10(sequencing_error_rate/3)) + min(prior_probs)
    # We need a heuristic to handle the fact that values below xmin return non-probabilities.
    e = min(1, powlaw_modified(x-y, powlaw_alpha, powlaw_xmin))
    return(e)


def _calculate_error_rate_table(max_hets, sequencing_error_rate, prior_probs, powlaw_alpha, powlaw_xmin):
    """
    Precompute empirical phasing error rates across a range of heterozygous count sites.
    Returns a list of error rates for heterozygous site counts from 0 to max_hets. These
    represent the error rates expected given a perfect match to a single phase, given a
    power law distribution with parameters powlaw_alpha and powlaw_xmin.
    """
    # This calculation is only correct for the diploid case!!

    # Numerator (all positions match)
    # For ploidy > 2, we would need to pick a specific phase and use the correct prior.
    x = np.array(range(1,max_hets+1))
    x = (x * np.log10(1 - sequencing_error_rate)) + max(prior_probs)
    
    # Denominator (all positions mismatch)
    # For ploidy > 2, this would need to be a loop over all other alleles,
    # with appropriate attention paid to using the correct prior for each.
    y = np.array(range(1,max_hets+1))
    y = (y * np.log10(sequencing_error_rate/3)) + min(prior_probs)

    # Convert to log-odds-ratios
    r = x - y

    # Extract error rates
    ret = []
    for i in range(max_hets):
        # We need a heuristic to handle the fact that values below xmin return non-probabilities.
        e = min(1, powlaw_modified(r[i], powlaw_alpha, powlaw_xmin))
        ret.append(e)
    # Handle zero case
    ret.insert(0,1)
    return(ret)

def error_analysis(args):
    """Main simulation mode

    Args:
        args:
    """    
    start_time = time.time()
    sys.stderr.write('\nPerforming error analysis using simulation mode...\n')
    sys.stderr.write('Analysis started %s.\n\n' % time.asctime())
    input_data = InputData(args.output_directory_path,
                           sample = args.one_sample,
                           ignore_samples = args.ignore_samples,
                           reference_sequence_input = args.reference_genome,
                           reference_sequence_input_assembly = args.reference_assembly,
                           max_threads = args.threads,
                           quiet = args.quiet_mode,
                           silent = args.silent_mode
    )
    input_data.add_haplotype_information(args.vcf_file_path)
    input_data.add_reference_sequence(args.reference_genome,
                                      reference_sequence_input_assembly = args.reference_assembly,
                                      output_directory = args.output_directory_path)

    if args.autosomal_only:
        # Only do simulations for autosomal sequences
        autosomal_ref_seq_path = create_fasta_file(
            input_fasta_file_path = args.reference_genome,
            output_fasta_file_path =  args.output_directory_path + '/reference_sequences/autosomal_' + args.reference_assembly + '_ref.fa',
            only_autosomal= True,
            chrom_sizes = args.chrom_sizes
        )
    else:
        autosomal_ref_seq_path = args.reference_genome
    
    """
    Some of what is implemented below is already contained in SimulatePhasedData.py and
    PhaseSet.py. However, I am choosing to reimplement here because I feel it will be faster
    and more efficient from a development standpoint to do so than to try and parse out how
    to use the object methods that Greg implemented. This should probably be addressed in
    a future release. --AGD
    """
    
    # Prepare the haploytpe-specific fasta
    haplotypes = [1,2]  # TO-DO: Extract the number of haplotypes from the VCF
    haplotype_specific_fasta = []
    chain_file_paths = []
    for haplotype in haplotypes:
        hsf, cfp = generate_haplotype_specific_fasta(
            haplotype,
            args.one_sample,
            autosomal_ref_seq_path,
            args.vcf_file_path,
            output_reference_sequence_path =
            args.output_directory_path +
            '/reference_sequences/autosomal_' +
            args.reference_assembly +
            '_ref_' +
            args.one_sample +
            '_hap' + str(haplotype) + '.fa',
            chain_file_path =
            args.output_directory_path +
            '/reference_sequences/autosomal_' +
            args.reference_assembly +
            '_ref_' +
            args.one_sample  +
            '_hap' + str(haplotype) + '.chain'
        )
        haplotype_specific_fasta.append(hsf)
        chain_file_paths.append(cfp)

    # Simulate reads at 1X depth for N iterations to achieve the necessary number of total reads.
    # Each 1X simulation is expected to produce ~ 190k reads.
    # TO-DO: - Simulate over all samples in VCF.
    #        - Make function args configurable through command-line options.
    #        - Add option to iterate until a minimum number of error observations
    #          are collected.
    hist_all = np.array([])
    hist_err = np.array([])
    bins = np.array([])
    error_lr_list = []
    error_read_list = []

    # Kludge to avoid needing a separate loop when args.simulation_min_err_obs is set:
    if args.simulation_min_err_obs > 0:
        args.simulation_depth = 9999999999999  # We just need an unreasonably large number of iterations to be safe!

    # Open a file for per-iteration simulation stats, if requested
    if args.store_simulation_data_for_iterations:
        per_iteration_stats_file = open(args.output_directory_path + "/per_iteration_simulation_error_stats.tsv", "w")
        
    sys.stderr.write("\n################ Simulating Error Data ################\n")
    for i in range(args.simulation_depth):
        sys.stderr.write("Starting iteration %d...\n" % (i+1))
        total_err_obs = len(error_lr_list)
        
        hist_all_i, hist_err_i, bins_i, error_lr_list_i, error_read_list_i = simulate_results_for_iteration(
            input_data,
            haplotypes,
            haplotype_specific_fasta,
            args
        )
        for j in range(len(error_lr_list_i)):
            error_lr_list.append(error_lr_list_i[j])
            error_read_list.append(error_read_list_i[j])
            
        if i > 0:
            # Combine with results of earlier iterations
            hist_all, hist_err, bins = combine_hists_and_bins(hist_all, hist_all_i,
                                                              hist_err, hist_err_i,
                                                              bins, bins_i,
                                                              args.log_likelihood_ratio_interval
            )
        else:
            hist_all, hist_err, bins = (hist_all_i, hist_err_i, bins_i)

        err_obs_i = len(error_lr_list) - total_err_obs
        sys.stderr.write("Iteration %d produced %d error observations.\n" % (i+1, err_obs_i))

        # Write cumulative error stats for each iteration to disk if requested
        if args.store_simulation_data_for_iterations:
            write_simulation_stats_for_iteration(hist_all, hist_err, bins, i, err_obs_i, len(error_lr_list), per_iteration_stats_file)
            
        # The rest of the kludge for using args.simulation_min_err_obs to control simulations:
        if args.simulation_min_err_obs > 0 and len(error_lr_list) >= args.simulation_min_err_obs:
            break

    # Close the per-iteration stats file if necessary.
    if args.store_simulation_data_for_iterations:
        close(per_iteration_stats_file)

    sys.stderr.write("\n################ Fitting Error Model ################\n")
    with open(args.output_directory_path + "/simulation_error_stats.tsv", "w") as outfile:
        outfile.write("log-odds-bin\ttotal-phased\ttotal-incorrect\teffective-error-rate\n")
        for i in range(len(hist_all)):
            err_rate = -1
            if hist_all[i] > 0:
                err_rate = hist_err[i]/hist_all[i]
            outfile.write("%.5f\t%d\t%d\t%.5f\n" % (bins[i], hist_all[i], hist_err[i], err_rate))
    # Fit the power-law model based on the observed log-likelihood-ratios and associated error rates.
    pl_fit = powerlaw.Fit(error_lr_list, method="KS")

    # Apply a linear xmin correction based on sample-size. 3.585 is the intercept term derived
    # from fitting a linear model to the difference between actual and estimated xmin values
    # over many sets of values sampled from powerlaw distributions with known parameters. We have
    # shown this correction to reduce the bias in xmin estimates across a wide range of sample sizes.
    corrected_xmin = ( pl_fit.xmin * (3.585 / np.log(len(error_lr_list))) ) + 3
    
    with open(args.output_directory_path + "/simulation_error_model_output.txt", 'w') as outfile:
        outfile.write("Total errors in simulations: %d\nMin log-likelihood-ratio for errors: %.3f\nMax log-likelihood ratio for errors: %.3f\nPowerlaw fitting method: %s\nNumber of observations used in powerlaw fitting: %d\nFitted powerlaw alpha: %.5f\nFitted powerlaw xmin: %.5f\nCorrected powerlaw xmin: %.5f\nFitted powerlaw minimum xmin distance (D): %.3f\nFitted powerlaw sigma: %.5f\n" % (len(error_lr_list), min(error_lr_list), max(error_lr_list), pl_fit.fit_method, pl_fit.n, pl_fit.alpha, pl_fit.xmin, corrected_xmin, pl_fit.D, pl_fit.sigma) )

    # Write the full error_lr_list to a file, if requested.
    if args.store_error_probs:
        with open(args.output_directory_path + "/error_lr_list.txt", 'w') as outfile:
            outfile.write("%s\n" % ",".join(error_lr_list))

    # Write detailed information on simulated error reads, if requested.
    if args.save_read_stats:        
        with open(args.output_directory_path + "/error_read_stats.txt", 'w') as outfile:
            outfile.write("seq_name\tlength\tseq_error_rate_avg\tn_hets\tphase\tn_matches_to_phase\tn_mismatches_to_phase\tlog_likelihood_hap1\tlog_likelihood_hap2\tlog_likelihood_ratio\tphasing_err_rate\n")
            for err_read in error_read_list:
                outfile.write("%s\t%d\t%.5f\t%d\t%d\t%d\t%d\t%.5f\t%.5f\t%.5f\t%.5f\n" % (
                    err_read.read_name,
                    err_read.aligned_segment.reference_length,
                    np.mean(err_read.PhaseSet_max.read_base_error_rates),
                    err_read.PhaseSet_max.total_hets_analyzed,
                    err_read.phase,
                    err_read.PhaseSet_max.matches,
                    err_read.PhaseSet_max.non_matches,
                    err_read.PhaseSet_max.log_probability_read_given_haplotype_i[0],
                    err_read.PhaseSet_max.log_probability_read_given_haplotype_i[1],
                    err_read.log_likelihood_ratio,
                    err_read.phasing_error_rate
                    )
                )

    runtime = time.time() - start_time
    sys.stderr.write('Analysis finished %s.\nTotal runtime %.2f seconds.\n' % (time.asctime(), runtime))
    return


def write_simulation_stats_for_iteration(hist_all, hist_err, bins, j, errs_j, errs_cumul, outfile):
    # Write simulation stats for iteration i to the given file.
    outfile.write("iteration: %d\n" % (j+1))
    outfile.write("errors for iteration: %d\tcumulative errors: %d\n" % (errs_j, errs_cumul))
    outfile.write("log-odds-bin\ttotal-phased\ttotal-incorrect\teffective-error-rate\n")
    for i in range(len(hist_all)):
            err_rate = -1
            if hist_all[i] > 0:
                err_rate = hist_err[i]/hist_all[i]
            outfile.write("%.5f\t%d\t%d\t%.5f\n" % (bins[i], hist_all[i], hist_err[i], err_rate))
    outfile.write("\n")
    

def combine_hists_and_bins(hist_all, hist_all_i, hist_err, hist_err_i, bins, bins_i, log_likelihood_ratio_interval):
    if len(bins_i) > len(bins):
        bins = np.append(bins, np.arange(max(bins),
                                         max(bins_i)+log_likelihood_ratio_interval,
                                         log_likelihood_ratio_interval)
        )
        hist_all = sum_unequal_hists(hist_all, hist_all_i)
        hist_err = sum_unequal_hists(hist_err, hist_err_i)
    elif len(bins) > len(bins_i):
        hist_all = sum_unequal_hists(hist_all_i, hist_all)
        hist_err = sum_unequal_hists(hist_err_i, hist_err)
    else:
        hist_all = hist_all + hist_all_i
        hist_err = hist_err + hist_err_i
        
    return hist_all, hist_err, bins


def sum_unequal_hists(shorter_arr, longer_arr):
    """
    Return the sum of two numpy histograms with different maximum bins.
    """
    return longer_arr + np.append(shorter_arr, np.zeros(len(longer_arr) - len(shorter_arr)))


def simulate_results_for_iteration(input_data, haplotypes, haplotype_specific_fasta, args):
    """
    Simulate data for a single iteration and tabulate results.
    """
    hist_all = np.array([])
    hist_err = np.array([])
    bins = np.array([])
    error_lr_list = []
    error_read_list = []
    for haplotype in haplotypes:
        sys.stderr.write("Simulating reads for haplotype %d...\n" % (haplotype))
        simulated_fastq = simulate_data(haplotype_specific_fasta[haplotype-1],
                                        args.simulation_error_model,
                                        args.output_directory_path+'/simulated_reads/',
                                        args.one_sample,
                                        haplotype,
                                        quiet = args.quiet_mode,
                                        silent = args.silent_mode)
        try:
            sys.stderr.write("Aligning simulated reads to reference genome...\n")
            input_data.add_reads(simulated_fastq,
                                 sample = args.one_sample,
                                 reference_sequence_input = args.reference_genome,
                                 clobber=True, quiet = args.quiet_mode,
                                 silent = args.silent_mode
            )
        except Exception as e:
            sys.stderr.write("%s\n" % e)
            continue

        
        # Extract log-likelihood ratios and phasing results for properly-mapped reads with >= min_het_sites
        sys.stderr.write("Phasing simulated reads...\n")
        hist_all_i, hist_err_i, bins_i, error_lr_list_i, error_read_list_i = parse_simulated_data(input_data, args)
        for j in range(len(error_lr_list_i)):
            error_lr_list.append(error_lr_list_i[j])
            error_read_list.append(error_read_list_i[j])
            
        if haplotype > 1:
            # Combine with results from other haplotype(s)
            hist_all, hist_err, bins = combine_hists_and_bins(hist_all, hist_all_i,
                                                              hist_err, hist_err_i,
                                                              bins, bins_i,
                                                              args.log_likelihood_ratio_interval
	    )
        else:
            hist_all, hist_err, bins = (hist_all_i, hist_err_i, bins_i)

        # Clean up after ourselves.
        os.remove(simulated_fastq)
        input_data._purge_alignment_files()
        sys.stderr.write("Done.\n")
        
    return hist_all, hist_err, bins, error_lr_list, error_read_list


def simulate_data(haplotype_specific_fasta, simulation_model,
                  output_directory, sample, haplotype,
                  depth = 1,
                  difference_ratio = '23:31:46', length_mean = 25000,
                  length_max = 1000000, length_min = 100, length_sd = 20000,
                  accuracy_min = 0.01, accuracy_max = 1.00,
                  accuracy_mean = 0.80,
                  quiet = False, silent = False):
    """
    Simulate reads for a single sample and haplotype.
    TO-DO: Make all optional args accessible as command-line options.
    """
    simulated_fastq = simulate_reads_pbsim2(
        reference_sequence = haplotype_specific_fasta,
        depth = depth,
        simulation_mode = simulation_model,
        difference_ratio = difference_ratio,
        length_mean = length_mean,
        length_max = length_max,
        length_min = length_min,
        length_sd = length_sd,
        accuracy_min = accuracy_min,
        accuracy_max = accuracy_max,
        accuracy_mean = accuracy_mean,
        output_directory = output_directory,
        sample = sample,
        haplotype = haplotype,
        quiet = quiet,
        silent = silent)
    
    return simulated_fastq


def parse_simulated_data(input_data, args):
    """
    Takes simulated alighments and parses out log-likelihood ratios and associated error counts.
    """
    log_ratios = []
    is_phased_correctly = []
    error_lr_list = []
    error_read_list = []
    for phasable_sample in input_data:
        # To-Do: Work in ability to select/ignore specific samples
        for alignment in phasable_sample:
            phased_read = PhasedRead(alignment, vcf_file = phasable_sample.vcf_file_path, sample = 'HG001', evaluate_true_alignment = True)
            if phased_read.alignment_is_mapped and phased_read.matches_true_alignment and phased_read.is_phased and int(phased_read.aligned_segment.get_tag('HS')) > 1:
                log_ratios.append(phased_read.log_likelihood_ratio)
                is_phased_correctly.append(phased_read.is_phased_correctly)
                if not phased_read.is_phased_correctly:
                    #sys.stderr.write("%s\n%s\n" % (phased_read, phased_read.is_phased_correctly))
                    error_lr_list.append(phased_read.log_likelihood_ratio)
                    error_read_list.append(phased_read)
        
    # Tabulate the total counts and number of errors at each log-odds level (binned in steps of bin_width).
    log_ratios = np.array(log_ratios)
    is_phased_correctly = np.array(is_phased_correctly, dtype=bool)
    bins = np.arange(0, np.ceil(max(log_ratios))+args.log_likelihood_ratio_interval, args.log_likelihood_ratio_interval)
    hist_all, edges = np.histogram(log_ratios, bins)
    hist_err, edges = np.histogram(log_ratios[np.nonzero(np.where(is_phased_correctly, is_phased_correctly == False, 1))], bins)
    
    return hist_all, hist_err, bins, error_lr_list, error_read_list


def getArgs() -> object:
    """Parses arguments:"""
    ################################################################################################################
    ############## LRphase Arguments ###############################################################################
    
    lrphase_parser: ArgumentParser = ArgumentParser(
        prog = 'LRphase', description = 'Tools for phasing individual long reads using haplotype information.'
    )
    
    lrphase_parser.add_argument('--version', action = 'version', version = __version__)

    lrphase_subparsers: _SubParsersAction = lrphase_parser.add_subparsers(
        title = '[LRphase modes]', dest = 'mode', description = 'Choose which mode to run:',
        help = 'mode must be added as first argument (ex: LRphase phasing)', required = True
    )

    
    ################################################################################################################
    ############## Phasing Mode Arguments ##########################################################################
    
    phasing_parser: ArgumentParser = lrphase_subparsers.add_parser(
        'phasing', description = "Tool for phasing individual long reads using haplotype information."
    )
    
    ############## Phasing Mode Required Arguments ##############
    phasing_parser_required: _ArgumentGroup = phasing_parser.add_argument_group('Required', 'Required for phasing')
    
    phasing_parser_required.add_argument(
        '-o', '--output_directory_name', required = True,
        help = 'Name given to directory where results will be output (ex: -o minion_GM12878_run3_phasing_output)',
        dest = 'output_directory_name', metavar = "</path/to/output>"
    )
    
    phasing_parser_required.add_argument(
        '-v', '--vcf', required = True,
        help = 'Path to vcf file with haplotype information that will be used for phasing. (Must be in .vcf.gz format '
        'with tabix index in same folder. If .vcf file is provided, bgzip and tabix must be installed and '
        'available on PATH because LRphase will attempt to convert it.  EX: -v GM12878_haplotype.vcf.gz)',
        dest = 'vcf_file_name', metavar = '<VCF_FILE>'
    )
    
    phasing_parser_required.add_argument(
        '-i', required = True,
        help = 'Path to alignment file (.bam or .sam) of long reads that will be used for phasing. If either a .sam '
        'file is provided or an index is not found, .sam and .bam file will be sorted and indexed using pysam. '
        'Sorted.bam files should be in same directory as their index (.sorted.bam.bai). EX: -a '
        'data/minion_GM12878_run3.sorted.bam, -i minion_GM12878_run3.sam) Path to long read file in .fastq '
        'format that will be used for alignment and phasing (ex: -i minion_GM12878_run3.fastq)',
        #dest = 'long_read_inputs', metavar = 'long-read input file', action = 'append', nargs = '*'
        dest = 'long_read_inputs', metavar = '<SAM/BAM/FASTQ>'
    )
    
    phasing_parser_required.add_argument(
        '-r', '--reference', required = False,
        help = 'Path to reference genome sequence file. REQUIRED if -i is used to specify reads in fastq format to be '
        'aligned prior to phasing. (file types allowed: .fa, .fna, fasta. EX: -r data/reference_hg38.fna)',
        dest = 'reference_genome', metavar = '<REF_FASTA>'
    )
    
    phasing_parser_required.add_argument(
        '-A', '--reference_assembly', required = False, default="hg38",
        help = 'Assembly for the reference genome. EX: -A hg38.',
        dest = 'reference_assembly', metavar = '<ASSEMBLY_NAME>'
    )

    phasing_parser_required.add_argument(
	'-t', '--threads', type=int, required=False, default=3,
        help = 'Number of threads to use for mapping step.',
        dest = 'threads', metavar = '<THREADS>'
    )
    
    ############## Phasing Mode Output Options ##############
    phasing_parser_output: _ArgumentGroup = phasing_parser.add_argument_group(
        'Output options', 'Options for tagging reads and printing output BAM files.'
    )

    phasing_parser_output.add_argument(
        '-H', '--omit_phasing_tag',
        help = 'Do not tag reads with "HP:i:1" (maternal) and "HP:i:2" (paternal) tags to indicate phase assignments.',
        dest = 'add_phasing_tag_to_output', action = 'store_false'
    )
    
    phasing_parser_output.add_argument(
        '-P', '--omit_phase_set_tag',
        help = 'Do not tag reads with "PS:i:x" tags, which label reads according to the phase set that was indicated in the vcf record used to assign a read to a phase.',
        dest = 'add_phase_set_tag_to_ouput', action = 'store_false'
    )

    phasing_parser_output.add_argument(
        '-Q', '--omit_phasing_quality_tag',
        help = 'Do not tag reads with "PC:i:x" tags, where x is an estimate of the accuracy of the phasing assignment in phred-scale.',
        dest = 'add_phasing_quality_tag_to_output', action = 'store_false'
    )

    phasing_parser_output.add_argument(
        '-O', '--output_mode', type=str, required=False, default="combined",
        choices=['combined', 'phase_tagged', 'full'],
        help = 'Specify whether/how phased, unphased, and nonphasable reads are printed to output. Modes available:\n\tcombined: All reads will be written to a common output file. The phasing tag (HP:i:N) can be used to extract maternal/paternal phased reads, unphased reads, and nonphasable reads.\n\tphase_tagged: Phased reads for both maternal and paternal phases will be written to a single output file, while unphased and nonphasable reads will be written to their own respective output files.\n\tfull: Maternal, paternal, unphased, and nonphasable reads will be printed to separate output files.',
        action = 'store'
    )
    
    ############## Multiple sample handling/phase set options for phasing mode ##############
    phasing_parser_multiple_sample = phasing_parser.add_argument_group(
        'Sample options',
        'Phasing options for files containing multiple samples or haplotypes. By default the input files are assumed '
        'to belong to the same haplotype and from the same sample. '
    )
    
    # phasing_parser_multiple_sample.add_argument('-R', '--add_read_group_tag_to_input_reads', required = False,
    # help='Use the --add_read_group_tag_to_input_reads option to overwrite or add @RG ID:Z:x SM:Z:y DS:Z:z headers
    # and RG:Z:x tags to all input reads where x is the ID number, y is the name of the sample to which the read
    # belongs, and z is a description of the experimental sample that generated the reads. (-R x y z)', nargs = 3,
    # dest = 'add_read_group_tag_to_input_reads')
    
    phasing_parser_multiple_sample.add_argument(
        '-s', '--one_sample', required = False,
        help = 'Use the --one_sample option to phase a specific sample present in the input reads and vcf file. (-s '
        'HG001)',
        metavar = '<SAMPLE_NAME>'
    )
    
    phasing_parser_multiple_sample.add_argument(
        '--ignore_samples',
        help = 'Use the --ignore_samples option to ignore sample labels. The first sample column in the VCF will be '
        'used and reads will not be matched using RG tags, samples, or phase sets.',
        action = 'store_true'
    )

    """
    # ############## Additional methods for evaluating phasing accuracy using simulated reads  ##############
    # Deprecated options? This functionality is now in error_analysis mode.

    # phasing_parser_simulation = phasing_parser.add_argument_group('Evaluation of phasing using simulated reads', 'Simulated reads are used to estimate the accuracy of phasing decisions and assign quality scores.')
    
    # phasing_parser_simulation.add_argument('--simulated', help='Use the --simulated option to enable extra analysis to be performed for determining correct phasing rates and selecting thresholds through iteration of phasing decisions over log likelihood thresholds 0-10. When --simulated is used, LRphase will skip phasing and process a per_read_phasing_stats.tsv results file located in the output folder specified by -o. ', action='store_true')
    
    # phasing_parser_simulation.add_argument('--simulated_bam', required = False, help = 'Path to file with phased reads to output results from simulated analyses.', dest = 'simulated_bam', metavar = 'phased output bam file')
    
    # phasing_parser_simulation.add_argument('--simulated_analysis_only', help='Use the --simulated option to enable extra analysis to be performed for determining correct phasing rates and selecting thresholds through iteration of phasing decisions over log likelihood thresholds 0-10.', action='store_true')
    
    # phasing_parser_simulation.add_argument('--simulated_phasing_stats', required = False, help = 'Path to file with phased reads to output results from simulated analyses.', dest = 'simulated_phasing_stats', metavar = 'phasing stats file')
    """
    
    ############## Statistical and error options for phasing mode ##############
    phasing_parser_stats_error: _ArgumentGroup = phasing_parser.add_argument_group(
        'Statistical options for phasing model',
        'Options to modify thresholds and error parameters involved in phasing decisions.'
    )

    phasing_parser_stats_error.add_argument(
        '--error_model', required = False, default = 0, choices = [0,1,2],
        help = 'Use the --error_model option to choose how to estimate sequencing error rates: 0: estimate per-base '
        'error rate as an average per read. 1: estimate per-base error rate locally around each het site. 2: '
        'Calculate per-base error rate using base quality scores (WARNING: do not use option 2 unless you are '
        'sure that the basecaller reported actual error rates). ',
        dest = 'error_model', type = int
    )

    """
    These options are disabled because the are not relevant to the current implementation of
    simulations through error_analysis mode.
    # phasing_parser_stats_error.add_argument('--read_average_error_model', help='Use the --read_average_error_model
    # option to estimate per-base error rate as an average per read.', action='store_false')
    
    #phasing_parser_stats_error.add_argument(
    #    '--error_stats',
    #    help = 'Use the --error_stats option to enable extra analysis to be performed for per-base error statistics.',
    #    action = 'store_true'
    #)
    
    #phasing_parser_stats_error.add_argument(
    #    '--phasing_error_file_path', required = False,
    #    help = "Path to file with estimated error statistics from simulated analyses.",
    #    dest = 'phasing_error_file_path', metavar = 'phasing_error_file_path'
    #)
    
    #phasing_parser_stats_error.add_argument(
    #    '--error_stats_hets', required = False,
    #    default = 'simulated_reads_log_likelihood_ratio_iteration_hets_phasing_stats.tsv',
    #    help = "'Path to file with estimated error statistics from simulated analyses.",
    #    dest = 'simulated_log_likelihood_ratio_iteration_hets_file', metavar = 'error statistics hets file'
    #)
    """

    phasing_parser_stats_error.add_argument(
        '-E', '--error_rate_threshold', type = float, required = False, default = 0.01,
        help = 'Error rate threshold on phasing results. This threshold equates to the estimated phasing error rate for an experiment, such that a threshold of 0.05 should be equivalent to a 5%% false-discovery rate. This rate is estimated based on a fitted power-law distribution with alpha and xmin parameters supplied with the --powlaw_alpha and --powlaw_xmin options. These parameters may be estimated by running the LRphase error_analysis mode.',
        dest = 'error_rate_threshold', metavar = '<MAX_ERROR_RATE>'
    )

    phasing_parser_stats_error.add_argument(
        '--log_likelihood_threshold', type = float, required = False, default = -1,
        help = 'Use a hard threshold on log-likelihood ratios instead of an error-rate threshold based on the fitted power-law distribution for error rates. Results will only be printed for predicted phasings with log-likelihood ratios equal to or greater than this threshold.',
        metavar = '<MIN_LIKELIHOOD_RATIO>'
    )

    phasing_parser_stats_error.add_argument(
        '--powlaw_alpha', type = float, required = False, default = 4.5,
        help = 'Alpha parameter for the fitted power law distribition used in error-rate calculations. This can be obtained by running LRphase error_analysis mode. Default = 4.5',
        dest = 'powlaw_alpha', metavar="<ALPHA>"
    )

    phasing_parser_stats_error.add_argument(
        '--powlaw_xmin', type = float, required = False, default = 2.0,
        help = 'Xmin parameter for the fitted power law distrubution used in error-rate calculations. This can be obtained by running LRphase error_analysis mode. Default = 2.0',
        dest = 'powlaw_xmin', metavar = '<XMIN>'
    )

    phasing_parser_stats_error.add_argument(
	'-q', '--quiet', help = 'Output to stderr from subprocesses will be muted.', action = 'store_true', dest = 'quiet_mode'
    )

    phasing_parser_stats_error.add_argument(
	'-S', '--silent', help = 'Output to stderr and stdout from subprocesses will be muted.', action = 'store_true', dest = 'silent_mode'
    )
    
    phasing_parser.set_defaults(func = phasing)


    #########################################################################################################################
    ############## Phasability Mode Arguments ###############################################################################
    
    phasability_parser = lrphase_subparsers.add_parser(
        'phasability',
        description = 'Tool that uses haplotype information as input and outputs predictions for how well LRphase will perform for a sequencing dataset with a given N50 and sequencing error rate. Phasability is defined as the probability of correctly phasing a read of width W that matches all variants in a single phase, calculated as (1 - empirical_phasing_error_rate) for any window of width W. Overlapping windows of width W are slid along the genome at intervals of step size S and results are reported in fixed-step wig format. Individual bins in the wig file describe phasability for the W-width window starting at the leftmost position of the bin.  Can be used to evaluate phasability genome-wide or at optional regions.'
    )
    
    ############## Phasability Mode Required Arguments ##############
    phasability_parser_required: _ArgumentGroup = phasability_parser.add_argument_group(
        'Required', 'Required for phasability analysis'
    )
    
    phasability_parser_required.add_argument(
        '-o', '--output_directory_name', required = True,
        help = 'Name given to directory where results will be output (ex: -o GM12878_phasability)',
        dest = 'output_directory_name', metavar = '</path/to/output>'
    )
    
    phasability_parser_required.add_argument(
        '-v', '--vcf', required = True,
        help = 'Path to vcf file with haplotype information that will be used for phasability analysis. (Must be in '
        '.vcf.gz format with tabix index in same folder. If .vcf file is provided, bgzip and tabix must be '
        'installed and available on PATH because LRphase will attempt to convert it. EX: -v '
        'GM12878_haplotype.vcf.gz)',
        dest = 'vcf_file_name', metavar = '<VCF_FILE>'
    )

    phasability_parser_required.add_argument(
        '-C', '--chrom_sizes', required = True, type=str,
        help = 'Chromosome sizes file. Required when using --autosomal_only',
        dest = 'chrom_sizes', metavar = '<CHROM_SIZES_FILE>'
    )
    
    ############## Phasability Mode Optional Arguments ##############
    phasability_parser_optional = phasability_parser.add_argument_group('Optional', 'Optional for phasability mode')
    
    phasability_parser_optional.add_argument(
        '-w', '--window_size', required = False, type=int, default = 25000,
        help = 'Window size for calculating phasability. Set this to a value near the N50 of the sequencing experiment you are evaluating. (default = 25000bp)',
        dest = 'window_size', metavar = '<N>'
    )
    
    phasability_parser_optional.add_argument(
        '-t', '--step_size', required = False, type = int, default = 1000,
        help = 'Distance between start positions of the overlapping windows used for phasability analysis (default = 1000bp)',
        dest = 'step_size', metavar = '<N>'
    )

    phasability_parser_optional.add_argument(
        '-e', '--sequencing_error_rate', required = False, type=float, default = 0.1,
        help = 'Estimaated per-base sequencing error rate. Typically ~0.1 for Nanopore sequencing. (Default = 0.1)',
        dest = 'sequencing_error_rate', metavar = '<ERROR_RATE>'
    )
    
    phasability_parser_optional.add_argument(
        '--powlaw_alpha', type = float, required = False, default = 4.5,
        help = 'Alpha parameter for the fitted power law distribition used in error-rate calculations. This can be obtained by running LRphase error_analysis mode. Default = 4.5',
        dest = 'powlaw_alpha', metavar="<ALPHA>"
    )

    phasability_parser_optional.add_argument(
        '--powlaw_xmin', type = float, required = False, default = 2.0,
        help = 'Xmin parameter for the fitted power law distrubution used in error-rate calculations. This can be obtained by running LRphase error_analysis mode. Default = 2.0',
        dest = 'powlaw_xmin', metavar = '<XMIN>'
    )
    
    phasability_parser_optional.add_argument(
        '-R', '--regions', required = False, type=str,
        help = 'Regions in which to estimate phasability, in BED format.',
        dest = 'regions', metavar = '<BED_FILE>'
    )
        
    phasability_parser_optional.add_argument(
        '-q', '--quiet', help = 'Output to stderr from subprocesses will be muted.', action = 'store_true', dest = 'quiet_mode'
    )

    phasability_parser_optional.add_argument(
	'-S', '--silent', help = 'Output to stderr and stdout from subprocesses will be muted.', action = 'store_true', dest = 'silent_mode'
    )
    
    phasability_parser.set_defaults(func = phasability)


    ##################################################################################################################
    ############## error_analysis Mode Arguments #####################################################################
    
    error_analysis_parser = lrphase_subparsers.add_parser(
        'error_analysis',
        description = 'Tool for estimating error rates in a dataset given a (set of) haplotype(s) and a reference genome. Simulated reads with known phase are generated based on the inputs, and processed through the LRphase phasing mode. Results are used to estimate parameters for the powerlaw distribution modeling empirical error rates within the data. (See https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0085777, and http://arxiv.org/abs/0706.1062 for more information.). Results of powerlaw model fittign are written to "simulation_error_model_output.txt" and a text report of error rates across the observed range of log-likelihood ratios is written to "simulation_error_stats.tsv" in the output directory. These files can be used to obtain the optimal parameters for setting thresholds for phasing mode: To use the fitted powerlaw model to set an error-rate based threshold (recommended), the "Fitted powerlaw alpha" and "Corrected powerlaw xmin" within "simulation_error_model_output.txt" represent the recommended values for the --powlaw_alpha and --powlaw_xmin options, respectively. This file also reports "Fitted powerlaw xmin", which is the actual xmin estimate from model fitting. This estimate tends to be systematically low, especially with fewer than 25,000 error observations, thus "Corrected powerlaw xmin" has been adjusted using a linear correction based on sample-size. If you are simulating more than 25,000 observations, you may wish to use the fitted value rather than the corrected value. To set a hard threshold on log-odds scores (NOT recommended), the values within "simulation_error_stats.tsv" can be used to build an eCDF from which a log-odds threshold can be chosen at any desired empirical error rate, which can be supplied to phasing mode using the --error_rate_threshold option.'
    )
    
    ############## error_analysis Mode Required Arguments ##############
    
    error_analysis_parser_required: _ArgumentGroup = error_analysis_parser.add_argument_group(
        'Required', 'Required for error_analysis'
    )
    
    error_analysis_parser_required.add_argument(
        '-o', '--output_directory_path', required = True,
        help = 'Name given to directory where results will be output (ex: -o minion_GM12878_run3_phasing_output)',
        dest = 'output_directory_path', metavar = '</path/to/output_directory>'
    )
    
    error_analysis_parser_required.add_argument(
        '-v', '--vcf', required = True,
        help = 'Path to vcf file with haplotype information that will be used for error_analysis. (Must be in .vcf.gz format with tabix index in same folder. If .vcf file is provided, bgzip and tabix must be installed and available on PATH because LRphase will attempt to convert it.  EX: -v GM12878_haplotype.vcf.gz)',
        dest = 'vcf_file_path', metavar = '<VCF_FILE>'
    )

    error_analysis_parser_required.add_argument(
        '-p', '--simulation_error_model', required = True,
        help = 'Path to the error model for simulated sequence generation',
        dest = 'simulation_error_model', metavar = '<ERROR_MODEL_FILE>'
    )
    
    error_analysis_parser_required.add_argument(
        '-r', '--reference', required = True,
        help = 'Path to reference genome sequence file.',
        dest = 'reference_genome', metavar = '<GENOME_FASTA>'
    )

    error_analysis_parser_required.add_argument(
        '-A', '--reference_assembly', required = False, default='hg38', type=str,
        help = 'Reference genome assembly, e.g., hg38.',
        dest = 'reference_assembly', metavar = '<ASSEMBLY_NAME>'
    )

    error_analysis_parser_required.add_argument(
        '-t', '--threads', type=int, required=False, default=3,
        help = 'Number of threads to use for mapping step.',
        dest = 'threads', metavar='<N>'
    )
    
    ############## Additional methods for evaluating phasing accuracy using simulated reads  ##############
    error_analysis_parser_simulation: _ArgumentGroup = error_analysis_parser.add_argument_group(
        'Evaluation of phasing using simulated reads',
        'Simulated reads are used to estimate the accuracy of phasing decisions and assign quality scores.'
    )

    error_analysis_parser_simulation.add_argument(
        '--ignore_phase_set',
        help = 'Compare the quality scores in a phased bam to error rates determined from simulated reads. All input '
        'must be phased bam files that  -i for ath to phased with phased reads to output results from '
        'simulated analyses.',
        action = 'store_true'
    )

    error_analysis_parser_simulation.add_argument(
        '-l', '--log_likelihood_ratio_interval', required = False, type=float, default = 0.5,
        help = 'Bin width for tracking phasing error at each log-likelihood-ratio level.',
        dest = 'log_likelihood_ratio_interval', metavar = '<N>'

    )

    error_analysis_parser_simulation.add_argument(
	'--simulation_depth', required = False, default=25, type=int,
        help = 'Total read depth coverage for all simulations.',
        dest = 'simulation_depth', metavar = '<N>'
    )

    error_analysis_parser_simulation.add_argument(
	'--simulation_min_err_obs', required = False, default=-1, type=int,
        help = 'Minimum number of error observations to require before stopping simulations. It is recommended that this be set >= 1000 for reliable estimation of the xmin parameter, though as few as 50 observations is sufficient to reliably estimate alpha if xmin is already known.',
        dest = 'simulation_min_err_obs', metavar = '<N>'
    )

    error_analysis_parser_simulation.add_argument(
        '--sample', required = False, type=str,
        help = 'Sample name in VCF to use for haploytpe simulation.',
        dest = 'one_sample', metavar = '<SAMPLE_NAME>'
    )

    error_analysis_parser_simulation.add_argument(
        '--ignore_samples', required = False,
        help = 'Ignore VCF samples. Simulations will be based on all samples.',
        dest = 'ignore_samples', action='store_true'
    )

    error_analysis_parser_simulation.add_argument(
        '--autosomal_only', required = False,
        help = 'Only simulate sequences on autosomes.',
        dest = 'autosomal_only', action='store_true'
    )

    error_analysis_parser_simulation.add_argument(
        '--chrom_sizes', required = False, type=str,
        help = 'Chromosome sizes file. Required when using --autosomal_only',
        dest = 'chrom_sizes', metavar = '<FILE>'
    )

    error_analysis_parser_simulation.add_argument(
        '--store_error_probs', required = False,
        help = 'Write the list of likelihood ratios for all error observations to a comma-delimited text file.',
        dest = 'store_error_probs', action='store_true'
    )

    error_analysis_parser_required.add_argument(
        '--save_detailed_read_stats', default=False,
        help = 'Save detailed information on simulated error reads. Data will be stored to <output_dir>/error_read_stats.txt',
        dest = 'save_read_stats', action = 'store_true'
    )

    error_analysis_parser_required.add_argument(
        '--store_simulation_data_for_iterations', default=False,
        help = 'Save simulated error stats for each iteration to disk. Data will be stored to <output_dir>/per_iteration_simulation_error_stats.tsv',
        dest = 'store_simulation_data_for_iterations', action = 'store_true'
    )

    error_analysis_parser_simulation.add_argument(
	'-q', '--quiet', help = 'Output to stderr from subprocesses will be muted.', action = 'store_true', dest = 'quiet_mode'
    )

    error_analysis_parser_simulation.add_argument(
	'-S', '--silent', help = 'Output to stderr and stdout from subprocesses will be muted.', action = 'store_true', dest = 'silent_mode'
    )
    
    """
    # Args not used in the current implementation...
    # error_analysis_parser_simulation.add_argument('--evaluate_phased_bam', help = 'Compare the quality scores in a
    # phased bam to error rates determined from simulated reads. All input must be phased bam files that  -i for ath
    # to phased with phased reads to output results from simulated analyses.', action='store_true')
    # error_analysis_parser_simulation.add_argument('--simulated_bam', required = False, help = 'Path to file with
    # phased reads to output results from simulated analyses.', dest = 'simulated_bam', metavar = 'phased output bam
    # file')
    # error_analysis_parser_simulation.add_argument('--simulated_analysis_only', help='Use the --simulated option to
    # enable extra analysis to be performed for determining correct phasing rates and selecting thresholds through
    # iteration of phasing decisions over log likelihood thresholds 0-10.', action='store_true')
    # error_analysis_parser_simulation.add_argument('--simulated_phasing_stats', required = False, help = 'Path to
    # file with phased reads to output results from simulated analyses.', dest = 'simulated_phasing_stats',
    # metavar = 'phasing stats file')
    #error_analysis_parser_simulation.add_argument(
    #    '--evaluate_phasing_stats',
    #    help = 'Compare the quality scores in a phased bam to error rates determined from simulated reads. All input '
    #    'must be phased bam files that  -i for ath to phased with phased reads to output results from '
    #    'simulated analyses.',
    #    action = 'store_true'
    #)
    #error_analysis_parser_simulation.add_argument(
    #    '-I', '--input_simulated', required = False,
    #    help = "Path to file with phased reads to output results from simulated analyses.",
    #    dest = 'input_simulated',
    #    metavar = 'phasing stats file', action = 'append', nargs = '*'
    #)
    #error_analysis_parser_simulation.add_argument(
    #    '-s', '--sample_output', required = False,
    #    help = "Path to file with phased reads to output results from simulated analyses.", dest = 'sample_output',
    #    metavar = '<FILE>'
    #)
    #error_analysis_parser_simulation.add_argument(
    #    '-P', '--phase_set_output', required = False, default = 'NA',
    #    help = 'Path to file with phased reads to output results from simulated analyses.',
    #    dest = 'phase_set_output',
    #    metavar = "<FILE>"
    #)
    #error_analysis_parser_simulation.add_argument(
    #    '-L', '--log_likelihood_ratio_threshold', required = False, default = 2.0,
    #    help = 'Path to file with phased reads to output results from simulated analyses.',
    #    dest = 'log_likelihood_ratio_threshold', metavar = 'phase set to analyze'
    #)
    #error_analysis_parser_simulation.add_argument(
    #    '-E', '--error_rate_threshold', required = False, default = 0.05,
    #    help = 'error rate used for phasing decision',
    #    dest = 'error_rate_threshold', type = float, metavar = 'error rate threshold'
    #)
    #error_analysis_parser_simulation.add_argument(
    #    '--phasing_error_file_path', required = False,
    #    help = 'Path to file with estimated error statistics from simulated analyses.',
    #    dest = 'phasing_error_file_path', metavar = 'phasing_error_file_path'
    #)
    """
    
    error_analysis_parser.set_defaults(func = error_analysis)

    
    #####################################################################################################################
    ############## Parse arguments ######################################################################################    
    args = lrphase_parser.parse_args()
    args.func(args)    
    return args


def main():
    # This only needs to call the getArgs method, which will then dispatch functions for the indicated runtime mode.
    args: Namespace = getArgs()
    
    
if __name__ == '__main__':
    main()



# coding=utf-8
import sys
import math
from collections import defaultdict
from statistics import mean
from typing import Any, List
import pysam
import numpy as np

import LRphase.HeterozygousSites as HeterozygousSite

# Minimum floating-point value > 0
# See http://www.seunosewa.com/2009/06/floating-point-min-max.html
minfloat = 4.9406564584124654e-324
log10minfloat = math.log10(minfloat)

def powlaw_modified(x: object, a: object = 4.5, xmin: object = 2) -> object:
    """

    Args:
        x:
        a:
        xmin:

    Returns:

    """
    return ((a - 1) / xmin) * math.pow(x / xmin, -1 * a)


"""
def _multcoeff(*args):
    '''
    # self.result = _multcoeff(*args)
    Return the multinomial coefficient
    (n1 + n2 + ...)! / n1! / n2! / ...
    '''
    if not args:  # no parameters
        return 1
    # Find and store the index of the largest parameter so we can skip
    #   it (for efficiency)
    skipndx = args.index(max(args))
    newargs = args[:skipndx] + args[skipndx + 1:]
    
    result = 1
    num = args[skipndx] + 1  # a factor in the numerator
    for n in newargs:
        for den in range(1, n + 1):  # a factor in the denominator
            result = result * num // den
            num += 1
    return result
"""

# Cache multinomial coefficients for efficiency
_multcoeff_cache = {}

def _multcoeff(n_match, n_mismatch):
    """
    Calculates multinomial coefficient according to the formula
    n! / ((n-m)! * m!)  (As described in the LRphase article)
    where n = the total number of heterozygous sites in a region 
    (n_match + n_mismatch) and m = the number of heterozygous sites
    matching the given haplotype (n_match). Values are cached to
    improve performance.
    """
    if (n_match, n_mismatch) not in _multcoeff_cache.keys():
        _multcoeff_cache[(n_match, n_mismatch)] = math.log10(np.math.factorial(n_match+n_mismatch) / (np.math.factorial(n_mismatch) * np.math.factorial(n_match)))
    return _multcoeff_cache[(n_match, n_mismatch)]
    

def _read_base_error_rates_from_quality_scores(gapped_alignment_positions, query_qualities):
    """# self.read_base_error_rates, self.error_rate_average_het_sites =
    self._read_base_error_rates_from_quality_scores(self.gapped_alignment_positions,
    self.aligned_segment.query_qualities) """
    
    ############ Extract base qualities for heterozygous sites in aligned_segment  ############
    
    read_base_error_rates = []
    error_rate_total_het_sites = 0
    total_het_sites_count = 0
    
    for position in gapped_alignment_positions:
        if isinstance(position, str):
            if position[0] == '-':
                read_base_error_rates.append(1)
        else:
            read_base_error_rates.append(10**(-0.1 * query_qualities[position]))
            error_rate_total_het_sites += 10**(-0.1 * query_qualities[position])
            total_het_sites_count += 1
    
    if total_het_sites_count > 0:
        error_rate_average_het_sites = error_rate_total_het_sites / total_het_sites_count
    else:
        error_rate_average_het_sites = 0
    return read_base_error_rates, error_rate_average_het_sites


def _read_bases_from_gapped_alignment(gapped_alignment_positions, query_sequence):
    """# self.read_bases = self._read_bases_from_gapped_alignment(self.gapped_alignment_positions,
    self.aligned_segment.query_sequence) """
    ############ Extract bases for heterozygous sites in aligned_segment  ############
    read_bases = []
    if query_sequence is None:
        return read_bases
    for position in gapped_alignment_positions:
        if isinstance(position, str):
            if position[0] == '-':
                read_bases.append('-')
        else:
            read_bases.append(query_sequence[position])
    return read_bases


def _read_base_error_rates_from_gapped_alignment(gapped_alignment_positions: object, per_base_error_rate):
    """# self.read_base_error_rates = self._read_base_error_rates_from_gapped_alignment(
    self.gapped_alignment_positions, self.per_base_error_rate) """
    read_base_error_rates = []
    # PhaseSet._estimate_per_base_error_rate()
    for position in gapped_alignment_positions:
        if isinstance(position, str):
            if position[0] == '-':
                read_base_error_rates.append(1)
        else:
            read_base_error_rates.append(per_base_error_rate)
    return read_base_error_rates


def _calculate_per_base_mismatch_rate(aligned_segment):
    """# self.per_base_mismatch_rate = self._calculate_per_base_mismatch_rate(self.aligned_segment)"""
    matches = 0
    i = 0
    for position in aligned_segment.get_aligned_pairs(with_seq = True):
        if position[1] is None:
            continue
        elif position[0] is None:
            continue
        else:
            if str(aligned_segment.query_sequence[position[0]]) == str(position[2]).upper():
                matches += 1
        i += 1
    per_base_mismatch_rate = float(1 - (matches / i))
    return per_base_mismatch_rate


def _calculate_local_per_base_mismatch_rate(aligned_segment, gapped_alignment_positions) -> object:
    """# self.read_base_error_rates, self.per_base_mismatch_rate, self.error_rate_average_het_sites =
    self._calculate_local_per_base_mismatch_rate(self.aligned_segment, self.gapped_alignment_positions)

    Args:
        aligned_segment:
        gapped_alignment_positions: """
    
    # read_base_error_rates = defaultdict(int)
    error_rate_total_het_sites = 0
    total_het_sites_count = 0
    matches = 0
    matches_alignment_query: List[int] = []
    read_base_error_rates = []
    # gapped_alignment_positions_error_rate = []
    i = 0
    for position in aligned_segment.get_aligned_pairs(with_seq = True):
        if position[1] is None:
            matches_alignment_query.append(0)
            continue
        elif position[0] is None:
            continue
        else:
            if str(aligned_segment.query_sequence[position[0]]) == str(position[2]).upper():
                matches += 1
                matches_alignment_query.append(1)
            else:
                matches_alignment_query.append(0)
        i += 1
    per_base_mismatch_rate = float(1 - (matches / i))
    
    for site in gapped_alignment_positions:
        if isinstance(site, str):
            if site[0] == '-':
                read_base_error_rates.append(1)
        else:
            try:
                read_base_error_rates.append(
                        round(
                                (2 * round(
                                        1 - (
                                                sum(matches_alignment_query[int(site) - 10:int(site) + 10]) / len(
                                                matches_alignment_query[int(site) - 10:int(site) + 10]
                                                )), 4
                                        ) + round(
                                        1 - (
                                                sum(matches_alignment_query[int(site) - 25:int(site) + 25]) / len(
                                                matches_alignment_query[int(site) - 25:int(site) + 25]
                                                )), 4
                                        ) + round(
                                        1 - (
                                                sum(matches_alignment_query[int(site) - 50:int(site) + 50]) / len(
                                                matches_alignment_query[int(site) - 50:int(site) + 50]
                                                )), 4
                                        )) / 4, 4
                                )
                        )
                error_rate_total_het_sites += round(
                        (2 * round(
                                1 - (
                                        sum(matches_alignment_query[int(site) - 10:int(site) + 10]) / len(
                                        matches_alignment_query[int(site) - 10:int(site) + 10]
                                        )), 4
                                ) + round(
                                1 - (
                                        sum(matches_alignment_query[int(site) - 25:int(site) + 25]) / len(
                                        matches_alignment_query[int(site) - 25:int(site) + 25]
                                        )), 4
                                ) + round(
                                1 - (
                                        sum(matches_alignment_query[int(site) - 50:int(site) + 50]) / len(
                                        matches_alignment_query[int(site) - 50:int(site) + 50]
                                        )), 4
                                )) / 4, 4
                        )
            except ZeroDivisionError:
                read_base_error_rates.append(per_base_mismatch_rate)
                error_rate_total_het_sites += per_base_mismatch_rate
            total_het_sites_count += 1
    
    if total_het_sites_count > 0:
        error_rate_average_het_sites = error_rate_total_het_sites / total_het_sites_count
    else:
        error_rate_average_het_sites = 0
    
    return read_base_error_rates, per_base_mismatch_rate, error_rate_average_het_sites


def _count_matches(
        read_bases: object, phased_alleles: object, read_base_error_rates: object,
        multinomial_correction: object
        ) -> object:
    """# self.log_probability_read_given_haplotype_i, self.non_matches, self.matches, self.total_hets,
    self.total_hets_analyzed, self.ploidy_phase_set = self._count_matches(self.read_bases, self.phased_alleles,
    self.read_base_error_rates, self.error_model)

    Args:
        read_bases:
        phased_alleles:
        read_base_error_rates:
        multinomial_correction:

    Returns:
        object: """
    log_probability_read_given_haplotype_i: defaultdict[Any, int] = defaultdict(int)  # log10(P(data|phase_i))
    non_matches = defaultdict(int)
    matches = defaultdict(int)
    total_hets = 0
    total_hets_analyzed = 0
    ploidy_list = []

    # print(read_bases, phased_alleles, read_base_error_rates, error_model)
    for i in range(len(read_bases)):
        read_base = read_bases[i]
        ploidy = len(phased_alleles[i])
        ploidy_list.append(ploidy)
        q = read_base_error_rates[i] + .001
        if not read_base == '-': # Skip gap positions
            """
            The loop below does not work consistently. In its current state, it is inconsistent
            in how complex variants are handled (i.e., those where one or more alleles is >1bp)
            in that these are counted as mismatches if the allele for the present haplotype is
            a single bp, but are skipped entirely when the current allele has len > 1. This means
            that log-likelihoods are calculated using different numbers of heterozygous sites per
            haplotype, which is incorrect. This commit adds functionality to filter complex variants
            out of the variant list prior to even running this step. This is done by default when the
            PhaseSet object is initialized. Given this change in behavior, the current implementation
            of this loop produces correct and consistent scores. However, this should be modified in
            future releases to calculate correct scores given all types of potential variants. In
            short, I'm leaving this as-is for now and relying on only simple SNPs being passed in
            in phased_alleles.
            """
            for haplotype in range(0, ploidy):
                # Begin relocated block
                if len(phased_alleles[i][haplotype]) > 1:  # Probably to exclude indels/rearrangements
                    continue
                total_hets_analyzed += 0.5  # These are counted twice
                # end relocated block
                """
                It is necessary to log-transform to avoid underflows. However, there may
                be a faster way to apply this that does not involve taking logs at every
                iteration!
                """
                if read_base == phased_alleles[i][haplotype]:
                    log_probability_read_given_haplotype_i[haplotype] += math.log10(1 - q)
                    matches[haplotype] += 1
                else:
                    log_probability_read_given_haplotype_i[haplotype] += math.log10(q / 3)
                    non_matches[haplotype] += 1
        total_hets += 1
    """
    sys.stderr.write("Total hets: %d, total_analyzed: %d, matches[0]: %d, mismatches[0]: %d, matches[1], %d, mismatches[1]: %d, q: %.5f\n" %
                     (total_hets, total_hets_analyzed,
                      matches[0], non_matches[0],
                      matches[1], non_matches[1], q)
    )
    """
    if total_hets_analyzed > 0:
        ploidy_phase_set = max(ploidy_list)
    else:
        ploidy_phase_set = 0
    
    if multinomial_correction:
        haplotype: int
        for haplotype in range(0, ploidy_phase_set):
            multinomial = _multcoeff(int(matches[haplotype]), int(non_matches[haplotype]))
            log_probability_read_given_haplotype_i[haplotype] = log_probability_read_given_haplotype_i[
                                                                    haplotype] + multinomial
            #sys.stderr.write("Haplotype %d log_prob: %.5f, prior: %.5f\n" % (haplotype, log_probability_read_given_haplotype_i[haplotype], multinomial))
    return log_probability_read_given_haplotype_i, non_matches, matches, total_hets, total_hets_analyzed, ploidy_phase_set


def _estimate_per_base_error_rate(aligned_segment: pysam.AlignedSegment) -> object:
    """# self.per_base_error_rate, self.error_rate_calculation = self._estimate_per_base_error_rate(
    self.aligned_segment) """
    error_rate_calculation = []
    try:
        per_base_error_rate = aligned_segment.get_tag(
                tag = 'de'
                )  # Gap-compressed per-base sequence divergence from minimap2 output
        error_rate_calculation.append('de tag')
    except LookupError:
        error_rate_calculation.append('no de tag')
        try:
            per_base_error_rate = aligned_segment.get_tag(tag = 'NM') / aligned_segment.query_alignment_length
            error_rate_calculation.append('NM tag')
        except LookupError:
            error_rate_calculation.append('no NM tag')
            try:
                per_base_error_rate = _calculate_per_base_mismatch_rate(aligned_segment)
                error_rate_calculation.append('_calculate_per_base_mismatch_rate success')
            except LookupError:
                error_rate_calculation.append('_calculate_per_base_mismatch_rate failed')
    if per_base_error_rate == 0:
        per_base_error_rate = 0.001
    return per_base_error_rate, error_rate_calculation


def _get_phased_alleles(vcf_records, sample, use_complex_variants = True, use_gap_variants = True):
    '''self.phased_alleles = self._get_phased_alleles(self.vcf_records, self.sample) '''
    provisional_recs = [vcf_record.samples[sample].alleles for vcf_record in vcf_records]

    if use_complex_variants and use_gap_variants:
        return provisional_recs
    elif len(provisional_recs) == 0:
        return provisional_recs

    haplotypes = range(len(provisional_recs[0]))
    ret = []
    for rec in provisional_recs:
        use_rec = True
        for haplotype in haplotypes:
            if not use_complex_variants and len(rec[haplotype]) > 1: # Drop complex variants
                use_rec = False
                break
            if not use_gap_variants and rec[haplotype] == '-': # Drop gap variants
                use_rec = False
                break
        if use_rec:
            ret.append(rec)

    return ret


def _get_reference_positions(vcf_records, sample,  use_complex_variants = True, use_gap_variants = True):
    '''self.reference_positions = self._get_reference_positions(self.vcf_records) '''
    provisional_pos = [vcf_record.pos - 1 for vcf_record in vcf_records] # Converts pos to zero-based coord. frame
    provisional_recs = [vcf_record.samples[sample].alleles for vcf_record in vcf_records]

    if use_complex_variants and use_gap_variants:
        return provisional_pos
    elif len(provisional_recs) == 0:
        return provisional_pos

    haplotypes = range(len(provisional_recs[0]))
    ret = []
    for i in range(len(provisional_recs)):
        rec = provisional_recs[i]
        use_rec = True
        for haplotype in haplotypes:
            if not use_complex_variants and len(rec[haplotype]) > 1:
                use_rec = False
                break
            if not use_gap_variants and rec[haplotype] == '-':
                use_rec = False
                break
        if use_rec:
            ret.append(provisional_pos[i])

    return ret


def _generate_gapped_alignment(reference_start, cigartuples, reference_positions) -> object:
    """# self.gapped_alignment_positions = self._generate_gapped_alignment(self.aligned_segment.reference_start,
    self.aligned_segment.cigartuples, self.reference_positions)

    Args:
        reference_start:
        cigartuples:
        reference_positions: """
    i: object = reference_start  # self.aligned_segment.reference_start
    j = 0
    h = 0
    gapped_alignment_positions = []
    
    if not cigartuples:
        return gapped_alignment_positions
    
    for cigartuple in cigartuples:  # self.aligned_segment.cigartuples:
        j1 = j
        if h >= len(reference_positions):
            break
        elif cigartuple[0] in [0, 7, 8]:
            i += cigartuple[1]
            j += cigartuple[1]
        elif cigartuple[0] in [1, 4]:
            j += cigartuple[1]
        elif cigartuple[0] in [2, 3]:
            i += cigartuple[1]
        while i > reference_positions[h]:
            if cigartuple[0] in [0, 7, 8]:
                gapped_alignment_positions.append(j1 + (cigartuple[1] - (i - reference_positions[h])))
                if h == len(reference_positions) - 1:
                    h += 1
                    break
            elif cigartuple[0] == 2:
                gapped_alignment_positions.append('-' + str(j1 + (cigartuple[1] - (i - reference_positions[h]))))
                if h == len(reference_positions) - 1:
                    h += 1
                    break
            if h >= len(reference_positions) - 1:
                break
            else:
                h += 1
    return gapped_alignment_positions


def _estimate_prior_probabilities(ploidy_phase_set = 2, prior_probabilities: object = None):
    """
    Returns log10 transformed priors.
    """
    if prior_probabilities is None:
        if ploidy_phase_set > 0:
            prior_probabilities = list(map(math.log10, [1 / ploidy_phase_set] * ploidy_phase_set))  # P(phase_i)
    return prior_probabilities


class PhaseSet:
    """

    """
    
    def __init__(
            self, phase_set, aligned_segment, vcf_records, sample, error_model = 0, error_rate_threshold = 0.01,
            multinomial_correction = True, prior_probabilities = None, liftover_converters = None, phase = 'Nonphasable',
            powlaw_alpha = 4.5, powlaw_xmin = 2, use_complex_variants = False, use_gap_variants = False
            ):
        self.aligned_segment = aligned_segment
        self.phase_set = phase_set
        self.vcf_records = vcf_records
        if liftover_converters:
            self.liftover_converters = liftover_converters
        else:
            self.liftover_converters = None
        # self.vcf_records
        self.sample = sample
        self.error_model = error_model
        self.error_rate_threshold = error_rate_threshold
        self.prior_probabilities = _estimate_prior_probabilities()  # Log10 transformed priors
        self.multinomial_correction = multinomial_correction
        self.gapped_alignment_positions = []
        self.read_bases = []
        self.phased_alleles = []
        self.reference_positions = []
        self.max_log_likelihood_ratio = None
        self.max_phase = None
        self.phase = phase
        self.powlaw_alpha = powlaw_alpha
        self.powlaw_xmin = powlaw_xmin
        
        self.use_complex_variants = use_complex_variants
        self.use_gap_variants = use_gap_variants
        
        self.total_hets = 0
        if not self.aligned_segment.is_unmapped: # or self.aligned_segment.is_aligned_to_contig_not_in_vcf:
            # Don't think second test is necessary because it should never yield len(self.vcf_records) > 0.
            # Current implementation of PhasedRead object sometimes leaves this attribute unset, causing
            # exceptions, so better if we can avoid it.
            if len(self.vcf_records) > 0:
                self.reference_positions = _get_reference_positions(self.vcf_records, self.sample, self.use_complex_variants, self.use_gap_variants)
                self.phased_alleles = _get_phased_alleles(self.vcf_records, self.sample, self.use_complex_variants, self.use_gap_variants)
                self.gapped_alignment_positions = _generate_gapped_alignment(
                        self.aligned_segment.reference_start,
                        self.aligned_segment.cigartuples,
                        self.reference_positions
                        )
                self.read_bases = _read_bases_from_gapped_alignment(
                        self.gapped_alignment_positions,
                        self.aligned_segment.query_sequence
                        )
        # self.read_base_error_rates =[]

        
    def __iter__(self):
        self.HeterozygousSites = []
        self.HeterozygousSites_processed_count = 0
        # self.alignment_files_read_counts = [bam_file.mapped+bam_file.unmapped for bam_file in self.bam_files]
        return self
    
    def __next__(self):
        if self.HeterozygousSites_processed_count < len(self.vcf_records):
            Heterozygous_Site = HeterozygousSite(
                    self.vcf_records[self.HeterozygousSites_processed_count], self.sample,
                    self.aligned_segment, self.liftover_converters
                    )
            self.HeterozygousSites.append(Heterozygous_Site)
            self.HeterozygousSites_processed_count += 1
            return Heterozygous_Site
        else:
            raise StopIteration()
    
    def solve_phase(
            self, error_model: object = 0, error_rate_threshold: object = 0.01, prior_probabilities: object = None,
            multinomial_correction: object = True
            ) -> object:
        """

        Args:
            error_model:
            error_rate_threshold:
            prior_probabilities:
            multinomial_correction:

        Returns:

        """
        if prior_probabilities is None:
            prior_probabilities = self.prior_probabilities
        if not self.aligned_segment.is_unmapped or self.aligned_segment.is_aligned_to_contig_not_in_vcf:
            if len(self.gapped_alignment_positions) > 0:
                if not error_model == 5:
                    self.read_base_error_rates, self.error_rate_average_het_sites, self.per_base_mismatch_rate = self._retrieve_read_base_error_rates(
                            self.aligned_segment, self.gapped_alignment_positions, error_model
                            )
                    
                    self.log_probability_read_given_haplotype_i, self.non_matches, self.matches, self.total_hets, self.total_hets_analyzed, self.ploidy_phase_set = _count_matches(
                            self.read_bases, self.phased_alleles, self.read_base_error_rates, multinomial_correction
                            )
                    
                    self.log_likelihood_ratios, self.phase, self.max_phase, self.max_log_likelihood_ratio = _phasing_decision(
                            self.log_probability_read_given_haplotype_i, self.ploidy_phase_set, prior_probabilities,
                            error_rate_threshold, self.total_hets_analyzed, powlaw_alpha=self.powlaw_alpha, powlaw_xmin=self.powlaw_xmin
                            )
                
                elif error_model == 5:
                    haplotypes = {}
                    self.matches = 0
                    self.non_matches = 0
                    hets = [het for het in self]
                    self.total_hets = len(hets)
                    self.log_likelihood_ratios = []
                    log_likelihood_ratio = {}
                    self.ploidy_phase_set = int(mean([int(het.ploidy_het_site) for het in self]))
                    for haplotype in range(1, self.ploidy_phase_set + 1):
                        haplotypes[haplotype] = sum(
                                [het.calculate_probs(haplotype = haplotype) for het in hets if
                                 het.calculate_probs(haplotype = haplotype)] # Second call to calculate_probs uses cached values
                                )
                        # if len(haplotypes[haplotype])>0:
                        # print(haplotypes[haplotype])
                    # print('1', haplotypes)
                    # print('2', haplotypes.values())
                    # print('3',list(haplotypes.values()))
                    
                    for count, haplotype in enumerate(haplotypes):
                        self.log_likelihood_ratios.append(haplotypes[haplotype] - haplotypes[len(haplotypes) - count])
                    self.max_log_likelihood_ratio = max(self.log_likelihood_ratios)
                    self.max_phase = self.log_likelihood_ratios.index(self.max_log_likelihood_ratio) + 1
                    if self.max_log_likelihood_ratio > self.error_rate_threshold:
                        self.phase = self.max_phase
                    else:
                        self.phase = 'Unphased'
                    self.total_hets = len(hets)
                    self.total_hets_analyzed = len(hets)
                    
                    self.log_probability_read_given_haplotype_i = haplotypes
                    for het in hets:
                        het.calculate_probs()
                        if het.total_matches > 0:
                            self.matches += 1
                        elif het.total_non_matches > 0:
                            self.non_matches += 1
                    
                    # self.matches = sum([1 for het in self if het.calculate_probs()[str(self.max_phase)][
                    # het.gapped_alignment_position]['match']]) self.non_matches = sum([1 for het in self if not
                    # het.calculate_probs()[str(self.max_phase)][het.gapped_alignment_position]['match']])
                    # per_base_mismatch_rate = aligned_segment.get_tag(tag='NM')/aligned_segment.query_alignment_length
                    # read_base_error_rates = [het.read_base_quality_error_rate for het in
                    # self]self._read_base_error_rates_from_gapped_alignment(gapped_alignment_positions,
                    # per_base_mismatch_rate)
        
        self.error_model_used = error_model
        self.error_rate_threshold_used = error_rate_threshold
        self.prior_probabilities_used = prior_probabilities
        
        return  # self.max_log_likelihood_ratio, self.phase_set
    
    def _retrieve_read_base_error_rates(
            self, aligned_segment: object, gapped_alignment_positions: object,
            error_model: object
            ) -> object:
        """

        Args:
            aligned_segment:
            gapped_alignment_positions:
            error_model:

        Returns:

        """
        error_rate_average_het_sites = None
        read_base_error_rates = []
        per_base_mismatch_rate = None
        
        if not self.aligned_segment.is_unmapped or self.aligned_segment.is_aligned_to_contig_not_in_vcf:
            if len(self.gapped_alignment_positions) > 0:
                
                if error_model == 3:
                    read_base_error_rates, per_base_mismatch_rate, error_rate_average_het_sites = _calculate_local_per_base_mismatch_rate(
                            aligned_segment, gapped_alignment_positions
                            )
                
                if error_model == 2:
                    read_base_error_rates, error_rate_average_het_sites = _read_base_error_rates_from_quality_scores(
                            gapped_alignment_positions, aligned_segment.query_qualities
                            )
                    per_base_mismatch_rate = aligned_segment.get_tag(
                            tag = 'NM'
                            ) / aligned_segment.query_alignment_length
                
                if error_model == 0:
                    per_base_mismatch_rate = aligned_segment.get_tag(tag = 'de')
                    read_base_error_rates = _read_base_error_rates_from_gapped_alignment(
                            gapped_alignment_positions, per_base_mismatch_rate
                            )
                
                if error_model == 1:
                    per_base_mismatch_rate = aligned_segment.get_tag(
                            tag = 'NM'
                            ) / aligned_segment.query_alignment_length
                    read_base_error_rates = _read_base_error_rates_from_gapped_alignment(
                            gapped_alignment_positions, per_base_mismatch_rate
                            )
                
                if error_model == 4:
                    per_base_mismatch_rate = _calculate_per_base_mismatch_rate(aligned_segment)
                    read_base_error_rates = _read_base_error_rates_from_gapped_alignment(
                            gapped_alignment_positions, per_base_mismatch_rate
                            )
                
                if len(read_base_error_rates) > 0:
                    if len([error_rate for error_rate in read_base_error_rates if error_rate != 1]) > 0:
                        error_rate_average_het_sites = mean(
                                [error_rate for error_rate in read_base_error_rates if error_rate != 1]
                                )
        
        return read_base_error_rates, error_rate_average_het_sites, per_base_mismatch_rate


def _phasing_decision(
        log_probability_read_given_haplotype_i, ploidy_phase_set, prior_probabilities,
        error_rate_threshold, total_hets_analyzed, powlaw_alpha = 4.5, powlaw_xmin = 2,
        minfloat = log10minfloat
        ) -> object:
    """# self.log_likelihood_ratios, self.phase, self.max_phase, self.max_log_likelihood_ratio =
    self._phasing_decision(self.log_probability_read_given_haplotype_i, self.ploidy_phase_set,
    self.prior_probabilities, self.error_rate_threshold)

    Args:
        log_probability_read_given_haplotype_i:
        ploidy_phase_set:
        prior_probabilities:
        error_rate_threshold:
        total_hets_analyzed: """
    
    if total_hets_analyzed == 0:
        phase = 'Nonphasable'
        log_likelihood_ratios = None
        max_log_likelihood_ratio: float = None
        max_phase = 'Nonphasable'
        return log_likelihood_ratios, phase, max_phase, max_log_likelihood_ratio
    
    else:
        log_likelihood_ratios = []
        for haplotype in log_probability_read_given_haplotype_i:
            denominator = 0
            numerator = 0
            for j in range(0, ploidy_phase_set):
                if haplotype == j:
                    # Assumes priors are log-transformed! 
                    numerator += (log_probability_read_given_haplotype_i[haplotype] +
                                  prior_probabilities[haplotype])
                    """
                    try:
                        # Assumes priors are log-transformed!
                        numerator += (log_probability_read_given_haplotype_i[haplotype] +
                                      prior_probabilities[haplotype])
                    except ValueError:
                        numerator += log_probability_read_given_haplotype_i[haplotype]
                    """
                else:
                    # Assumes priors are log-transformed! 
                    denominator += 10**(log_probability_read_given_haplotype_i[j] + prior_probabilities[j])
            #sys.stderr.write("hap: %d, numer: %f, denom %f\n" % (haplotype, numerator, denominator))
                    
            # This heuristic is needed to account for edge cases in which a read
            # has a very large number of heterozygous sites, most of which match
            # a single phase. These can cause underflows in the denominator.
            try:
                denom = math.log10(denominator)
            except ValueError:
                denom = log10minfloat
            log_likelihood_ratios.append(numerator - denom)

        #sys.stderr.write("%s\n" % (log_likelihood_ratios))
        max_log_likelihood_ratio = max(log_likelihood_ratios)
        if max_log_likelihood_ratio == 0:
            phase = 'Unphased'
            max_phase = 0
        else:
            max_phase = log_likelihood_ratios.index(max_log_likelihood_ratio) + 1
            if powlaw_modified(max_log_likelihood_ratio, a=powlaw_alpha, xmin=powlaw_xmin) <= error_rate_threshold:
                # if max_log_likelihood_ratio >= 10 ** (math.log10((error_rate_threshold / .5491)) / -2.45):
                phase = max_phase
            else:
                phase = 'Unphased'
        
        return [log_likelihood_ratios, phase, max_phase, max_log_likelihood_ratio]



# coding=utf-8
import time
from typing import Any, Iterator
import sys, os
import pysam
from LRphase import SimulatePhasedData
import pyliftover


class PhasableSample:
    """

    """
    #RG_ID_dict: object 
    
    def __init__(
            self,
            sample,
            vcf_file_path,
            ignore_phase_sets,
            alignment_file_paths,
            RG_ID_dict,
            bam_header = None,
            reference_sequence_names = None,
            reference_sequence_paths = None,
            simulate_haplotypes = False,
            output_directory_path = None,
            only_autosomal = False,
            regions = None,
            ploidy = 2,
            quiet = False,
            silent = False
    ):
        self.sample = sample
        
        self.vcf_file = pysam.VariantFile(vcf_file_path)
        self.vcf_file_path = vcf_file_path
        self.ignore_phase_sets = ignore_phase_sets
        
        self.alignment_file_paths = []
        self.RG_ID_dict = RG_ID_dict
        self.use_RG_tag = {}
        
        #sys.stderr.write("%s\n" %(alignment_file_paths))        
        for alignment_file_path in alignment_file_paths:
            self.alignment_file_paths.append(str(alignment_file_path))
            self.use_RG_tag[str(alignment_file_path)] = alignment_file_paths[str(alignment_file_path)]

        if bam_header:
            self.bam_header = bam_header
        else:
            # Take the header from the first alignment file.
            self.bam_header = pysam.AlignmentFile(self.alignment_file_paths[0], 'rb').header

        self.reference_sequence_names = reference_sequence_names
        self.reference_sequence_paths = reference_sequence_paths
        self.reference_sequences_in_VCF = list(self.vcf_file.index)
        self.only_autosomal = only_autosomal
        self.regions = regions
        self.ploidy = ploidy
        self.quiet = quiet
        self.silent = silent
        
        if output_directory_path is None:
            self.output_directory = str(self.sample)
        else:
            self.output_directory = output_directory_path
        if not os.path.exists(self.output_directory):
            os.mkdir(self.output_directory)
        
        if simulate_haplotypes:
            
            self.reference_sequence_path = str(os.path.abspath(self.output_directory)) + '/reference_sequence.fa'
            reference_sequence_list = []
            for i, reference_sequence in enumerate(self.reference_sequence_paths):
                reference_sequence_list.append(SimulatePhasedData.create_fasta_file(input_fasta_file_path = reference_sequence, output_fasta_file_path = str(os.path.abspath(self.output_directory)) + '/reference_sequence'+str(i)+'.fa', regions = self.regions, only_autosomal = self.only_autosomal))
            
            with open(self.reference_sequence_path, 'a') as outfile:
                for file in reference_sequence_list:
                    with open(file, 'r') as infile:
                        outfile.write(infile.read())
                        
            SimulatePhasedData.index_fasta_file(self.reference_sequence_path)
            
            self.haplotype_reference_sequences = {}
            self.haplotype_chain_files = {}
            self.haplotype_liftover_converters = {}
            for j in range(1,self.ploidy+1):
                self.haplotype_reference_sequences[j], self.haplotype_chain_files[j] = SimulatePhasedData.generate_haplotype_specific_fasta(j, self.sample, self.reference_sequence_path, self.vcf_file_path, output_reference_sequence_path = str(os.path.abspath(self.output_directory)) + '/haplotype'+str(j)+'_reference_sequence.fa', chain_file_path = str(os.path.abspath(self.output_directory)) + '/haplotype'+str(j)+'_reference_sequence.chain')
                self.haplotype_liftover_converters[j] = pyliftover.LiftOver(self.haplotype_chain_files[j])

    def _pysam_bam_files_initialize(self):
        bam_files = []
        for alignment_file_path in self.alignment_file_paths:
            bam_files.append(iter(pysam.AlignmentFile(alignment_file_path, 'rb')))
        return bam_files

    def __iter__(self):
        #sys.stderr.write('Alignments for sample %s\n' % str(self.sample))
        self.bam_files = self._pysam_bam_files_initialize()
        # self._initialize_alignment_counter()
        self.alignment_files_processed_count = 0
        self.alignments_processed_count = 0
        self.alignment_files_read_counts = []
        bam_file: Iterator[Any]
        for bam_file in self.bam_files:
            self.alignment_files_read_counts.append(bam_file.mapped + bam_file.unmapped)
        self.total_alignments = sum(self.alignment_files_read_counts)
        self.alignment_file_pysam = self.bam_files[self.alignment_files_processed_count]
        self.start_process_time = time.time()
        self.unique_read_names = set()
        return self

    def __next__(self):
        read = next(self.alignment_file_pysam)
        if read.query_name not in self.unique_read_names:
            self.unique_read_names.add(read.query_name)
        #read = self._evaluate_alignment(read)
        RG_info: object = self._get_RG_info_for_read(read, self.alignment_file_pysam)
        #if str(RG_info[2])[0:3] == 'SIM':
        #read.set_tag(tag='oa', value=str(RG_info[2][3]), value_type='Z', replace=True)
        read.set_tag(tag = 'RG', value = str(RG_info[1]), value_type = 'Z', replace = True)
        self.alignment_files_read_counts[self.alignment_files_processed_count] -= 1
        if self.alignment_files_read_counts[self.alignment_files_processed_count] == 0:
            self.alignment_files_processed_count += 1
            if self.alignment_files_processed_count == len(self.alignment_file_paths):
                raise StopIteration()
            self.alignment_file_pysam = self.bam_files[self.alignment_files_processed_count]
        self.alignments_processed_count += 1
        if self.alignments_processed_count % 2000 == 0 and not (self.quiet or self.silent):
            print(
                    'Processing %s reads per second' % str(
                            round(self.alignments_processed_count / (time.time() - self.start_process_time), 2)
                            )
                    )
            print(
                    'Processed %s reads (%s percent complete)' % (str(self.alignments_processed_count), str(
                            round(100 * (self.alignments_processed_count / self.total_alignments), 2)
                            ))
                    )
            print(
                    'Processing sample %s will finish in %s seconds' % (str(self.sample), str(
                            round(
                                    (self.total_alignments - self.alignments_processed_count) / (
                                            self.alignments_processed_count / (time.time() - self.start_process_time)),
                                    1
                                    )
                            ))
                    )
        return read

    def __repr__(self):
        return f'Sample Name: {self.sample}\nVCF: {self.vcf_file_path}\nReference sequence path: {str(self.reference_sequence_paths)}\nTotal sequences in reference files: {len(self.reference_sequence_names)}\nTotal Reference sequences in VCF: {len(self.reference_sequences_in_VCF)}\nAlignment Files: {str(self.alignment_file_paths)}\nTotal alignment files processed: {self.alignment_files_processed_count}\nTotal alignments: {self.total_alignments}\nTotal alignments processed: {self.alignments_processed_count}\nTotal unique reads observed: {len(self.unique_read_names)}'

    def _get_RG_info_for_read(self, read: pysam.AlignedSegment, alignment_file: object) -> object:
        """

        Args:
            read: 
            alignment_file: 

        Returns:

        """
        alignment_file_path = ''
        if isinstance(alignment_file, pysam.AlignmentFile):
            if str(alignment_file.filename.decode()) in self.alignment_file_paths:
                alignment_file_path = str(alignment_file.filename.decode())
            else:
                return
        elif str(alignment_file) in self.alignment_file_paths:
            alignment_file_path = str(alignment_file)
        else:
            return
        if alignment_file_path in self.use_RG_tag.keys():
            use_RG_tag = self.use_RG_tag[alignment_file_path]
            if use_RG_tag:
                if str(read.get_tag('RG')) in self.RG_ID_dict[alignment_file_path]:
                    if self.RG_ID_dict[alignment_file_path][str(read.get_tag('RG'))]['SM'] == self.sample:
                        ID = str(read.get_tag('RG'))
                        outputID = self.RG_ID_dict[alignment_file_path][str(read.get_tag('RG'))]['outputID']
                        sample_description = self.RG_ID_dict[alignment_file_path][str(read.get_tag('RG'))]['DS']
                        return ID, outputID, sample_description, use_RG_tag
            else:
                ID = list(self.RG_ID_dict[alignment_file_path].keys())[0]
                outputID = self.RG_ID_dict[alignment_file_path][ID]['outputID']
                sample_description = self.RG_ID_dict[alignment_file_path][ID]['DS']
                return ID, outputID, sample_description, use_RG_tag

            
    def simulate_reads(self, path_to_pbsim = 'pbsim', depth = 1,
        simulation_mode = 'pbsim2/data/R103.model',
        difference_ratio = '23:31:46', length_mean = 20000,
        length_max = 1000000, length_min = 100,
        length_sd = 15000, accuracy_min = 0.01,
        accuracy_max = 1.00, accuracy_mean = 0.80,
        prefix = None, id_prefix = 'S', output_directory = None, sample = None,
        haplotypes = [1,2]):
        
        self.haplotype_simulated_reads = {}
        for haplotype in haplotypes:
            self.haplotype_simulated_reads[haplotype] = SimulatePhasedData.simulate_reads_pbsim2(self.haplotype_reference_sequences[haplotype],
                                                                                                 path_to_pbsim = path_to_pbsim,
                                                                                                 depth = depth,
                                                                                                 simulation_mode = simulation_mode,
                                                                                                 difference_ratio = difference_ratio,
                                                                                                 length_mean = length_mean,
                                                                                                 length_max = length_max,
                                                                                                 length_min = length_min,
                                                                                                 length_sd = length_sd,
                                                                                                 accuracy_min = accuracy_min,
                                                                                                 accuracy_max = accuracy_max,
                                                                                                 accuracy_mean = accuracy_mean,
                                                                                                 prefix = prefix,
                                                                                                 id_prefix = id_prefix,
                                                                                                 output_directory = self.output_directory +
                                                                                                 '/simulated',
                                                                                                 sample = self.sample,
                                                                                                 haplotype = haplotype
            )
            
            
    
        
                       
        
    # @classmethod
    # def PhasableSample(
    #         cls, sample, vcf_file_path, ignore_phase_sets, param, RG_ID_dict: object, reference_sequence_names,
    #         reference_sequence_paths
    #         ):
    #     """
    #
    #     Args:
    #         sample:
    #         vcf_file_path:
    #         ignore_phase_sets:
    #         reference_sequence_names:
    #         reference_sequence_paths:
    #         RG_ID_dict (object):
    #
    #     Returns:
    #         object:
    #     """
    #     pass



# coding=utf-8
import sys, os
import random
import subprocess
import time
from collections import defaultdict
from datetime import date
from shutil import copyfileobj
from typing import Any, Dict, Tuple, List, Optional, Union

import pysam
import requests
from LRphase import PhasableSample
from LRphase.PhasedRead import *
from pysam import VariantFile
from LRphase import urls


def _prepare_output_directory(output_directory_path: str) -> object:#Optional[str]:
    """

    Args:
        output_directory_path (object): 
    """
    if os.path.isfile(output_directory_path):
        sys.stderr.write('Output directory exists as a file. Use -o to specify the name to be given to the output folder. May also provide a relative path with a name to create output directory in a different location (EX: -o path/to/name). Do not specify a file.\n')
        return
    elif os.path.exists(output_directory_path):
        sys.stderr.write('%s already exists, WARNING: New results will overwrite old files.\n' % os.path.abspath(output_directory_path))
    elif not os.path.exists(output_directory_path):
        os.mkdir(output_directory_path)
        sys.stderr.write("%s created\n" % os.path.abspath(output_directory_path))
    output_directory = os.path.abspath(output_directory_path)
    return output_directory


def _prepare_summary_file(output_directory):
    sys.stderr.write('\n############## Preparing summary_file ##############\n')
    
    summary_file_path = os.path.abspath('%s/%s' % (output_directory, 'summary.txt'))
    if os.path.exists(summary_file_path):
        sys.stderr.write('%s already exists, WARNING: New results will overwrite old files.\n' % os.path.abspath(summary_file_path))
    elif not os.path.exists(summary_file_path):
        sys.stderr.write('%s created.\n' % os.path.abspath(summary_file_path))
    summary_file = open(summary_file_path, 'w')
    summary_file.write('Start date: %s\n' % str(date.today()))
    start_time = time.time()
    sys.stderr.write('%s created.\n' % summary_file_path)
    
    return summary_file_path


def _pair_sample_with_vcf(sample, sample_to_vcf_file_dict, ignore_phase_sets):
    if str(sample) in sample_to_vcf_file_dict.keys():
        for vcf_file_path in sample_to_vcf_file_dict[str(sample)].items():
            if ignore_phase_sets:
                return vcf_file_path[0], True
            else:
                return vcf_file_path[0], vcf_file_path[1]
    else:
        return


def _sample_to_alignment_files(sample_to_vcf_file_dict: dict, RG_ID_dict: dict) -> object:
    """
    Create a dictionary relating sample names to alignment files.
    """
    sample_to_alignment_files = {}
    for sample in sample_to_vcf_file_dict:
        #sys.stderr.write("%s\n" % sample)
        sample_to_alignment_files[sample] = {}
        for file in RG_ID_dict.keys():
            #sys.stderr.write("%s\n" % file)
            if len([file for key in RG_ID_dict[file].keys() if RG_ID_dict[file][key]['SM'] == sample]) > 0:
                #sys.stderr.write("%s\n" % file)
                for pair in [{
                        alignment_file_path:any(
                                [RG_ID_dict[alignment_file_path][ID]['RG_tags'] for ID in
                                 RG_ID_dict[alignment_file_path].keys()]
                                )
                        } for alignment_file_path in RG_ID_dict.keys() if any(
                        [RG_ID_dict[alignment_file_path][ID]['SM'] == sample for ID in
                         RG_ID_dict[alignment_file_path].keys()]
                        )]:
                    #sys.stderr.write("%s\n" % pair)
                    sample_to_alignment_files[sample][list(pair.keys())[0]] = list(pair.values())[0]
    return sample_to_alignment_files


def _align_long_reads_fastq(long_reads_fastq_path, reference_sequence_input, output_directory, threads, quiet = False, silent = False):
    """
    Align specified reads file to reference genome via minimap2.
    """
    if not reference_sequence_input:
        sys.stderr.write('no reference genome was provided for %s. This file will be skipped.\n' % long_reads_fastq_path)
        return
    
    sys.stderr.write('\n################ Beginning Sequence Alignment ################\n')
    
    start_process_time = time.time()
    reference_sequence_input_path = os.path.abspath(reference_sequence_input)
    if str(long_reads_fastq_path).lower().endswith('.fastq'):
        sam_alignment_path = os.path.abspath(
                '%s/%s%s' % (
                        output_directory, os.path.splitext(os.path.basename(long_reads_fastq_path))[0],
                        '_alignment.sam')
                )
    elif str(long_reads_fastq_path).lower().endswith('fastq.gz'):
        sam_alignment_path = os.path.abspath(
                '%s/%s%s' % (
                        output_directory,
                        os.path.splitext(os.path.splitext(os.path.basename(long_reads_fastq_path))[0])[0],
                        '_alignment.sam')
                )

    mapping_cmd_args = ['minimap2', '-ax', 'map-ont', '-Y', '-L', '--secondary=no',
                        '--MD', reference_sequence_input_path,
                        long_reads_fastq_path,
                        '-o', sam_alignment_path,
                        '-t', str(threads)]
    if silent:
        subprocess.run(mapping_cmd_args, check = True, stderr = subprocess.DEVNULL, stdout = subprocess.DEVNULL)
    elif quiet:
        subprocess.run(mapping_cmd_args, check = True, stderr = subprocess.DEVNULL)
    else:
        subprocess.run(mapping_cmd_args, check = True)


    
    end_process_time = time.time()
    total_process_time = end_process_time - start_process_time
    
    sys.stderr.write('Alignment finished in %.2f seconds.\n' % total_process_time)
    
    return sam_alignment_path


def _sort_and_index_alignment_file(long_reads_alignment_path, output_directory: object):
    start_process_time = time.time()
    
    sorted_bam_file_path = os.path.abspath(
            '%s/%s%s' % (
                    output_directory, os.path.splitext(os.path.basename(long_reads_alignment_path))[0],
                    '_sorted.bam')
            )
    sys.stderr.write('\n############## Sorting and indexing bam file ##############\n')
    pysam.sort('-O', 'BAM', long_reads_alignment_path, '-o', sorted_bam_file_path)
    sys.stderr.write('Created %s.\n' % sorted_bam_file_path)
    pysam.index(sorted_bam_file_path)
    sys.stderr.write('Created %s.bai.\n' % sorted_bam_file_path)
    os.remove(long_reads_alignment_path)
    end_process_time = time.time()
    total_process_time = end_process_time - start_process_time
    sys.stderr.write('Sorting and indexing finished in %.2f seconds.\n' % total_process_time)
    return sorted_bam_file_path


def _prepare_alignment(output_directory: str, long_reads_alignment_path: str) -> str:
    start_process_time = time.time()
    long_read_file_pysam = pysam.AlignmentFile(long_reads_alignment_path)
    if long_read_file_pysam.format == 'BAM':
        if long_read_file_pysam.has_index():
            sorted_bam_file_path = os.path.abspath(long_reads_alignment_path)
            sys.stderr.write('%s is a valid alignment file with an index.\n' % long_reads_alignment_path)
        elif not long_read_file_pysam.has_index():
            sys.stderr.write('%s is a .bam file but the index cannot be found. Sorting and indexing bam file.\n' % long_reads_alignment_path)
            sorted_bam_file_path = _sort_and_index_alignment_file(long_reads_alignment_path, output_directory)
    elif long_read_file_pysam.format == 'SAM':
        sys.stderr.write('%s is a .sam file. Converting to binary (.bam), sorting, and indexing bam file.\n' % long_reads_alignment_path)
        sorted_bam_file_path = _sort_and_index_alignment_file(long_reads_alignment_path, output_directory)
    else:
        sys.stderr.write('Error: Pysam does not recognize %s as being in SAM or BAM format. If aligned reads are provided as input they must be in proper .sam or .bam format.' % long_reads_alignment_path)
        raise Exception('Improper alignment file format.')
    return sorted_bam_file_path


def _unique_RG_IDs_from_RG_tags(RG_ID_dict: dict, unique_RG_IDs: dict, alignment_file_path: str) -> object:
    """

    Args:
        RG_ID_dict (dict):
        unique_RG_IDs (dict):
        alignment_file_path (str):
    """
    with pysam.AlignmentFile(alignment_file_path, 'rb') as bam_file:
        RG_tags = bam_file.header.get('RG')
    if RG_tags is None:
        RG_ID_dict[str(alignment_file_path)] = 'No RG tags'
    else:
        RG_ID_dict[str(alignment_file_path)] = {}
        for RG_tag in RG_tags:
            RG_ID_dict[str(alignment_file_path)][str(RG_tag['ID'])] = {}
            RG_ID_dict[str(alignment_file_path)][str(RG_tag['ID'])]['DS'] = str(RG_tag['DS'])
            RG_ID_dict[str(alignment_file_path)][str(RG_tag['ID'])]['SM'] = str(RG_tag['SM'])
            RG_ID_dict[str(alignment_file_path)][str(RG_tag['ID'])]['RG_tags'] = True
            if str(RG_tag['ID']) in list(unique_RG_IDs):
                if unique_RG_IDs[str(RG_tag['ID'])]['DS'] == str(RG_tag['DS']) and unique_RG_IDs[str(RG_tag['ID'])][
                    'SM'] == str(RG_tag['SM']):
                    RG_ID_dict[str(alignment_file_path)][str(RG_tag['ID'])]['outputID'] = str(RG_tag['ID'])
                else:
                    not_unique = True
                    i = 0
                    while not_unique:
                        newID: str = str(RG_tag['ID']) + '_' + str(i)
                        if str(newID) not in list(unique_RG_IDs):
                            RG_ID_dict[str(alignment_file_path)][str(RG_tag['ID'])]['outputID'] = str(newID)
                            unique_RG_IDs[str(newID)] = {}
                            unique_RG_IDs[str(newID)]['DS'] = str(RG_tag['DS'])
                            unique_RG_IDs[str(newID)]['SM'] = str(RG_tag['SM'])
                            not_unique = False
                        i += 1
            else:
                RG_ID_dict[str(alignment_file_path)][str(RG_tag['ID'])]['outputID'] = str(RG_tag['ID'])
                unique_RG_IDs[str(RG_tag['ID'])] = {}
                unique_RG_IDs[str(RG_tag['ID'])]['DS'] = str(RG_tag['DS'])
                unique_RG_IDs[str(RG_tag['ID'])]['SM'] = str(RG_tag['SM'])
    return RG_ID_dict, unique_RG_IDs


def _extract_RG_info_from_long_read_input(long_read_input: List[str]) -> object:#tuple[Union[Union[str, bytes], Any], object, Optional[Any], Optional[Any], Optional[Any]]:
    """

    Returns:
        object:
    """
    long_read_input_path = os.path.abspath(long_read_input[0])
    if len(long_read_input) == 1:
        input_ID = None
        input_sample = None
        input_sample_description = None
        input_reference_sequence_input = None
    elif len(long_read_input) == 2:
        input_ID = long_read_input[1]
        input_sample = None
        input_sample_description = None
        input_reference_sequence_input = None
    elif len(long_read_input) == 3:
        input_ID = long_read_input[1]
        input_sample = long_read_input[2]
        input_sample_description = None
        input_reference_sequence_input = None
    elif len(long_read_input) == 4:
        input_ID = long_read_input[1]
        input_sample = long_read_input[2]
        input_sample_description = long_read_input[3]
        input_reference_sequence_input = None
    elif len(long_read_input) >= 5:
        input_ID: object = long_read_input[1]
        input_sample = long_read_input[2]
        input_sample_description = long_read_input[3]
        input_reference_sequence_input = long_read_input[4]
    return long_read_input_path, input_ID, input_sample, input_sample_description, input_reference_sequence_input


def _sample_to_vcf_file_dict(vcf_file_paths):
    """
    Create a dictionary relating sample names to vcf file paths.
    """
    sample_to_vcf_file_dict = {}
    for vcf_file_path in vcf_file_paths:
        vcf_file = pysam.VariantFile(list(vcf_file_path.keys())[0])
        for sample in vcf_file.header.samples:
            sample_to_vcf_file_dict[str(sample)] = vcf_file_path
        vcf_file.close() # Clean up after ourselves!
    return sample_to_vcf_file_dict


def _compile_read_groups(
        alignment_file_path, sample, ID, sample_description, RG_ID_dict, unique_RG_IDs, ignore_samples
        ):
    if ignore_samples:
        RG_ID_dict[str(alignment_file_path)] = 'ignore_samples'
    else:
        with pysam.AlignmentFile(alignment_file_path, 'rb') as bam_file:
            RG_tags = bam_file.header.get('RG')
        if sample is not None:
            RG_ID_dict[str(alignment_file_path)] = {}
            if ID is not None:
                RG_ID_dict[str(alignment_file_path)][str(ID)] = {}
                RG_ID_dict[str(alignment_file_path)][str(ID)]['SM'] = str(sample)
                RG_ID_dict[str(alignment_file_path)][str(ID)]['outputID'] = str(ID)
                RG_ID_dict[str(alignment_file_path)][str(ID)]['RG_tags'] = False
                if sample_description is not None:
                    RG_ID_dict[str(alignment_file_path)][str(ID)]['DS'] = str(sample_description)
                else:
                    RG_ID_dict[str(alignment_file_path)][str(ID)]['DS'] = str(
                            'LRphase_input_file_' + str(alignment_file_path)
                            )
            else:
                ID = '0' + str(random.randint(1, 10000))
                RG_ID_dict[str(alignment_file_path)][str(ID)] = {}
                RG_ID_dict[str(alignment_file_path)][str(ID)]['SM'] = str(sample)
                RG_ID_dict[str(alignment_file_path)][str(ID)]['outputID'] = str(ID)
                RG_ID_dict[str(alignment_file_path)][str(ID)]['RG_tags'] = False
                if sample_description is not None:
                    RG_ID_dict[str(alignment_file_path)][str(ID)]['DS'] = str(sample_description)
                else:
                    RG_ID_dict[str(alignment_file_path)][str(ID)]['DS'] = str(
                            'LRphase_input_file_' + str(alignment_file_path)
                            )
            if str(ID) not in unique_RG_IDs:
                unique_RG_IDs[str(ID)] = {}
                unique_RG_IDs[str(ID)]['DS'] = str(sample_description)
                unique_RG_IDs[str(ID)]['SM'] = str(sample)
        elif RG_tags is not None:
            RG_ID_dict, unique_RG_IDs = _unique_RG_IDs_from_RG_tags(
                    RG_ID_dict, unique_RG_IDs, alignment_file_path
                    )
        else:
            sys.stderr.write("%s Has No RG tags and was not input with sample information. Reads in this file will not be processed. Either re-input this read file with sample information or resubmit with ignore_samples option.\n" % alignment_file_path)
    return RG_ID_dict, unique_RG_IDs


class InputData:
    """
    sample_to_RG_header: defaultdict[Any, list]
    sample_to_PG_header: defaultdict[Any, list]
    unique_RG_IDs: Dict[Any, Any]
    """
    output_directory: object
    try:
        from LRphase import urls
        urls_found = True
    except:
        sys.stderr.write('Could not find import urls from data. Data will not be able to downloaded from example web sources.\n')
        urls_found = False
    
    def __init__(
            self, output_directory_path: str = None, vcf_file_input: str = None, long_read_input: object = None,
            reference_sequence_input: str = None, sample: str = None, ID: str = None,
            sample_description: str = None,
            ignore_phase_sets: bool = None, ignore_samples: bool = None, download_from_urls: bool = False,
            reference_sequence_input_assembly: str = None, auto_simulate_samples: bool = False,
            max_threads: int = 3, quiet = False, silent = False
            ) -> object:
        """

        Returns:
            object: 
        """
        sys.stderr.write('\n############## Preparing output directory tree ##############\n')
        if output_directory_path is not None:
            self.output_directory = _prepare_output_directory(output_directory_path)
        else:
            self.output_directory = _prepare_output_directory(
                    'LRphase_output_' + str(time.localtime()[0]) + '-' + str(time.localtime()[1]) + '-' + str(
                            time.localtime()[2]
                            ) + '_' + str(time.localtime()[3]) + 'hr_' + str(time.localtime()[4]) + 'min_' + str(
                            time.localtime()[5]
                            ) + 'sec'
                    )
        
        _prepare_output_directory(self.output_directory + '/reference_sequences')
        _prepare_output_directory(self.output_directory + '/haplotype_information')
        _prepare_output_directory(self.output_directory + '/output_reads')
        _prepare_output_directory(self.output_directory + '/input_reads')
        
        self.summary_file_path = _prepare_summary_file(self.output_directory)
        self.reference_sequence_paths = None
        self.RG_ID_dict = {}
        self.unique_RG_IDs = {}
        self.vcf_files = []
        self.sample_to_vcf_file_dict = {}
        self.sample_to_reference_sequences_dict = defaultdict(list)
        self.sample_to_reference_sequence_path = defaultdict(list)
        self.sample_to_sam_header = defaultdict(list)
        self.sample_to_PG_header = defaultdict(list)
        self.sample_to_RG_header = defaultdict(list)
        self.alignment_file_to_reference_sequences_dict = defaultdict(list)
        self.alignment_files = []
        self.phasable_samples = {}
        self.reference_sequence_files = {}
        
        self.auto_simulate_samples = auto_simulate_samples
        
        # self.sample_hap_to_true_alignment_dict = {}
        
        self.reference_sequence_input = reference_sequence_input
        self.vcf_file_input = vcf_file_input
        self.long_read_input = long_read_input
        self.sample = sample
        self.ID = ID
        self.sample_description = sample_description
        self.ignore_phase_sets = ignore_phase_sets
        self.ignore_samples = ignore_samples
        self.download_from_urls = download_from_urls
        self.reference_sequence_input_assembly = reference_sequence_input_assembly

        self.max_threads = max_threads
        self.quiet = quiet
        self.silent = silent

        # To store log-likelihood ratios and associated error rates from simulations.
        self.lr_err_rate_dict = {}
        
        if not self.reference_sequence_input_assembly is None:
            sys.stderr.write('\n############## Preparing reference sequence directory ##############\n')
            _prepare_output_directory(
                    self.output_directory + '/reference_sequences/' + str(reference_sequence_input_assembly)
                    )

        if not self.long_read_input is None:
            # self.long_read_input = long_read_input
            long_read_input_path, input_ID, input_sample, input_sample_description, input_reference_sequence_input = _extract_RG_info_from_long_read_input(long_read_input)
            if input_reference_sequence_input is None:
                input_reference_sequence_input = self.reference_sequence_input
            self.add_reads(self.long_read_input, input_sample, input_ID,
                           input_sample_description, input_reference_sequence_input)
        
    
    def _download_file(self, url, output_directory = None):
        if output_directory is None:
            local_filename = self.output_directory + '/' + url.split('/')[-1]
        else:
            local_filename = output_directory + '/' + url.split('/')[-1]
        with requests.get(url, stream = True) as r:
            with open(local_filename, 'wb') as f:
                copyfileobj(r.raw, f)
        return local_filename
    

    # Removed commented functions. -- AGD
    

    def add_reference_sequence(
            self, reference_sequence_input, output_directory = None, reference_sequence_input_assembly = None
            ) -> object:
        """
        Create an pysam.FastaFile object and store it within self.reference_sequence_files[].
        First downloads the reference sequence if supplied as a URL.
        If fasta file is bgzipped, unzips it.
        Stores absolute path to the reference sequence file in self.reference_sequence_files[].
        Returns absolute path to current reference sequence file.
        """
        #sys.stderr.write("%s; %s; %s\n" % (reference_sequence_input, output_directory, reference_sequence_input_assembly))

        # First, build the directory path where the reference sequence file will be written.
        if output_directory is None:
            output_directory_path = self.output_directory
        else:
            output_directory_path = output_directory
        
        if reference_sequence_input_assembly is None:
            output_directory_path = self.output_directory + '/reference_sequences'
        else:
            output_directory_path = self.output_directory + '/reference_sequences/' + str(reference_sequence_input_assembly)
            sys.stderr.write('\n############## Preparing reference sequence directory ##############\n')
            _prepare_output_directory(output_directory_path)

        # Download the reference sequence if a URL was provided. Reassign reference_sequence_input
        # to hold the location + name of the downloaded file.
        if reference_sequence_input.startswith('http'):
            sys.stderr.write('The reference sequence input is a url. Downloading now.\n')
            reference_sequence_input = self._download_file(reference_sequence_input, output_directory_path)

        # So this looks like it just reassembles the path to the given input file
        # and applies the .fa extension. Could we just test for different file types
        # instead?
        #reference_sequence_input_path = os.path.dirname(os.path.abspath(reference_sequence_input)) + '/' + os.path.basename(reference_sequence_input).split('.')[0] + '.fa'
        reference_sequence_input_path = '.'.join(reference_sequence_input.split('.')[0:-1]) + '.fa'  # Does same thing with less ops

        # Instantiate the pysam.FastaFile object given the supplied/downloaded file. Assign
        # this object to self.reference_sequence_files, indexed by the file path.
        try:
            ref_seq = pysam.FastaFile(reference_sequence_input_path)
        except (OSError, FileNotFoundError, IsADirectoryError, PermissionError) as e:
            # There was a problem accessing the reference sequence file.
            sys.stderr.write("ERROR: Reference sequence file {} is absent or unreadable.\nError message was: {}\n".format(reference_sequence_input_path, e))
            exit(2)

        # Only attempt to decode if we have a valid, readable file.
        try:
            reference_sequence_file_path: object = ref_seq.filename.decode()
            self.reference_sequence_files[reference_sequence_file_path] = ref_seq
            if reference_sequence_input_assembly:
                self.reference_sequence_files[reference_sequence_input_assembly] = ref_seq

        # If that fails, try unzipping the file with bgzip before instantiating the pysam.FastaFile
        # and making the subsequent value assignments.
        except OSError:
            # This appears to make the implicit assumption that an OSError means
            # that reference_sequence_input is a bgzip file -- maybe not always true??
            # Also, all this does is decompress the input file. Is there a way to
            # avoid this necessity? -- AGD
            sys.stderr.write("ERROR: {}\n".format(OSError))
            with pysam.BGZFile(reference_sequence_input, 'r') as infile:
                with open(reference_sequence_input_path, 'w') as outfile:
                    outfile.write(infile.read().decode())
            ref_seq = pysam.FastaFile(reference_sequence_input_path)
            #reference_sequence_names = ref_seq.references
            reference_sequence_file_path: object = ref_seq.filename.decode()
            self.reference_sequence_files[reference_sequence_file_path] = ref_seq
            if reference_sequence_input_assembly:
                self.reference_sequence_files[reference_sequence_input_assembly] = ref_seq

        # Removed large block of commented code. --AGD

        # Store the reference sequence path to self.reference_sequence_paths[].
        if self.reference_sequence_paths is None:
            self.reference_sequence_paths = [reference_sequence_file_path]
        elif isinstance(self.reference_sequence_paths, str):
            _reference_sequence_paths = []
            _reference_sequence_paths.append(self.reference_sequence_paths)
            self.reference_sequence_paths = _reference_sequence_paths
            self.reference_sequence_paths.append(reference_sequence_file_path)
        elif isinstance(self.reference_sequence_paths, list):
            self.reference_sequence_paths.append(reference_sequence_file_path)
        sys.stderr.write('Reference sequence prepared successfully: %s\n' % reference_sequence_file_path)

        # Return the current reference sequence path.
        return reference_sequence_file_path  #reference_sequence_names

    
    def add_haplotype_information(
            self, vcf_file_input: str, ignore_phase_sets: bool = False, reference_sequence: object = 'hg38'
            ) -> object:
        """
        
        """
        # Check vcf_input_file for proper format and indexing
        vcf_file_path = self._prepare_vcf_file(self.output_directory, vcf_file_input)

        # Check for phasing information
        if pysam.VariantFile(vcf_file_path).header.formats.keys().count('PS') == 0:
            ignore_phase_sets = True
            sys.stderr.write("This VCF file does not have the PS subfield. Phase sets will be ignored and all phased variants on the same chromosome (vcf contig) will be considered to be one contiguous haploblock " % vcf_file_path)

        # Append the VCF file path to self.vcf_files[], etc.
        self.vcf_files.append({vcf_file_path:ignore_phase_sets})
        self.sample_to_vcf_file_dict = _sample_to_vcf_file_dict(self.vcf_files)
        self.sample_to_alignment_files = _sample_to_alignment_files(self.sample_to_vcf_file_dict, self.RG_ID_dict)
        self._sample_to_reference_sequences_dict()
        self._sample_to_PG_dict()

        # Return path to the vcf file.
        return vcf_file_path

    
    def __iter__(self):
        self.sample_counter = 0
        self.phasable_samples = {}
        return self
    
    def new_PhasableSample(self, sample: str, reference_sequence_paths: list = None, auto_simulate_samples: bool = False) -> object:
        """

        Args:
            sample:
            reference_sequence_paths:

        Returns:

        """
        vcf_file_path, ignore_phase_sets = _pair_sample_with_vcf(
                sample, self.sample_to_vcf_file_dict, self.ignore_phase_sets
                )
        reference_sequence_names: list = self.sample_to_reference_sequences_dict[sample]
        #sys.stderr.write("%s\n" % sample)
        #sys.stderr.write("%s\n" % self.sample_to_alignment_files[sample])
        phasable_sample = PhasableSample.PhasableSample(
            sample, vcf_file_path, ignore_phase_sets, self.sample_to_alignment_files[sample], self.RG_ID_dict,
            reference_sequence_names, reference_sequence_paths = reference_sequence_paths,
            simulate_haplotypes = auto_simulate_samples, output_directory_path=self.output_directory,
            quiet = self.quiet, silent = self.silent
        )
        return phasable_sample
    
    def __next__(self):
        if self.sample_counter < len(list(self.sample_to_vcf_file_dict.keys())):
            sample = str(list(self.sample_to_vcf_file_dict.keys())[self.sample_counter])
            if sample in self.sample_to_reference_sequence_path:
                if len(self.sample_to_reference_sequence_path[sample]) > 0:
                    reference_sequence_paths = self.sample_to_reference_sequence_path[sample]
            else:
                reference_sequence_paths = self.reference_sequence_paths
            phasable_sample = self.new_PhasableSample(sample, reference_sequence_paths, self.auto_simulate_samples)
            #vcf_file_path, ignore_phase_sets = self._pair_sample_with_vcf(sample, self.sample_to_vcf_file_dict,
            #                                                             self.ignore_phase_sets)
            #phasable_sample = PhasableSample(sample, vcf_file_path, ignore_phase_sets,
            #                               self.sample_to_alignment_files[sample], self.RG_ID_dict)
            self.sample_counter += 1
            self.phasable_samples[phasable_sample.sample] = phasable_sample
            return phasable_sample
        else:
            raise StopIteration()
    
    def add_reads(
            self, long_reads_alignment_path, sample = None, ID = None, haplotype = None, sample_description = None,
            reference_sequence_input = None, database = None, master_database = False, simulated = False,
            reference_sequence_input_assembly = None, preserve_alignment = False, clobber = False,
            quiet = False, silent = False
            ):
        """
        clobber arg will cause any existing reads/alignments to be purged prior to loading current read set when True.
        """
        if clobber:
            self._reinit_reads(preserve_alignments=preserve_alignment)
        
        if long_reads_alignment_path.startswith('http'):
            sys.stderr.write('The reads input is a url. Downloading now.\n')
            long_reads_alignment_path = self._download_file(long_reads_alignment_path, self.output_directory)

        try:
            sorted_bam_file_paths, combined_long_read_fastq_path = self._parse_long_reads_input(long_reads_alignment_path,
                                                                                                self.output_directory)
        except Exception as e:
            sys.stderr.write("Error parsing long-read file: %s\n" % e.strerror)
            exit(2)
        if combined_long_read_fastq_path:
            if reference_sequence_input is None and self.reference_sequence_input is None:
                sys.stderr.write("Must supply reference sequence!\n")
                raise Exception("Reference sequence not given.")
            elif self.reference_sequence_input is None:
                reference_sequence_input = self.add_reference_sequence(
                    reference_sequence_input = reference_sequence_input,
                    output_directory = self.output_directory,
                    reference_sequence_input_assembly = reference_sequence_input_assembly
                )
            else:
                reference_sequence_input = self.reference_sequence_input
                
            sorted_bam_file_paths.append(
                _align_long_reads_fastq(
                    combined_long_read_fastq_path, reference_sequence_input, self.output_directory, self.max_threads, quiet = quiet, silent = silent
                )
            )
        for _sorted_bam_file_path in sorted_bam_file_paths:
            if _sorted_bam_file_path:
                sorted_bam_file_path, combined_long_read_fastq_path = self._parse_long_reads_input(
                        _sorted_bam_file_path, self.output_directory
                        )
                #sys.stderr.write("%s; %s\n" % (sorted_bam_file_path, combined_long_read_fastq_path))
                self.RG_ID_dict, self.unique_RG_IDs = _compile_read_groups(
                    sorted_bam_file_path[0], sample, ID, sample_description,
                    self.RG_ID_dict, self.unique_RG_IDs,
                    self.ignore_samples
                )
                #sys.stderr.write("%s; %s\n" % (self.RG_ID_dict, self.unique_RG_IDs))
                self.alignment_files.append(sorted_bam_file_path[0])
                #sys.stderr.write("%s\n" % self.alignment_files)
                self.sample_to_alignment_files = _sample_to_alignment_files(
                        self.sample_to_vcf_file_dict, self.RG_ID_dict
                        )
                #sys.stderr.write("%s\n" % self.sample_to_alignment_files)
                self._sample_to_reference_sequences_dict(sorted_bam_file_path[0])
                self._sample_to_PG_dict(sam_file_path = sorted_bam_file_path[0])
                if reference_sequence_input is not None:
                    if not reference_sequence_input in self.sample_to_reference_sequence_path[sample]:
                        self.sample_to_reference_sequence_path[sample].append(reference_sequence_input)


    def _reinit_reads(self, preserve_alignments=True):
        """
        Reinitialize all objects related to storing reads in an InputData objects.
        """
        self.RG_ID_dict = {}
        self.unique_RG_IDs = {}
        if not preserve_alignments:
            self._purge_alignment_files()
        self.alignment_files = []
        self.sample_to_reference_sequences_dict = defaultdict(list)
        self.sample_to_reference_sequence_path = defaultdict(list)
        self.sample_to_PG_header = defaultdict(list)
        self.sample_to_RG_header = defaultdict(list)
        return
    

    def _purge_alignment_files(self):
        """
        Purge alignment files from the filesystem.
        """
        for alignment_file in self.alignment_files:
            if os.path.isfile(alignment_file):
                os.remove(alignment_file)
            if os.path.exists("%s.bai" % (alignment_file)):
                os.remove("%s.bai" % (alignment_file))            
        self.alignment_files = []
        return

    
    def _prepare_sam_header(self, _sample = None):
        if not str(_sample) in self.sample_to_sam_header:
            self.sample_to_sam_header[_sample] = defaultdict(list)
        self.sample_to_sam_header[_sample]['HD'] = {'VN':'1.6', 'SO':'coordinate'}
        self.sample_to_sam_header[_sample]['SQ'] = self.sample_to_reference_sequences_dict[_sample]
        self.sample_to_sam_header[_sample]['RG'] = self.sample_to_RG_header[_sample]
        self.sample_to_sam_header[_sample]['PG'] = self.sample_to_PG_header[_sample]
        return  # self.sample_to_sam_header[_sample]
    

    def _sample_to_reference_sequences_dict(self, sam_file_path = None, sam_file = None):
        """
        Prepare a dictionary relating reference sequences to samples.
        """
        if sam_file_path:
            aln_file = pysam.AlignmentFile(sam_file_path)
            for reference_sequence in aln_file.header['SQ']:
                self.alignment_file_to_reference_sequences_dict[sam_file_path].append(reference_sequence)
            aln_file.close()
        for sample in self.sample_to_vcf_file_dict:
            for sam_file_path in self.alignment_file_to_reference_sequences_dict:
                if sam_file_path in self.sample_to_alignment_files[str(sample)]:
                    aln_file = pysam.AlignmentFile(sam_file_path)
                    for reference_sequence in aln_file.header['SQ']:
                        if str(reference_sequence['SN']) not in [str(reference_seq['SN']) for reference_seq in
                                                                 self.sample_to_reference_sequences_dict[sample]]:
                            self.sample_to_reference_sequences_dict[sample].append(reference_sequence)
                    aln_file.close()
                    
    
    def _sample_to_PG_dict(self, sam_file_path: object = None, sam_file: object = None) -> object:
        """
        Creates a dictionary relating samples to phasing groups.
        Args:
            sam_file_path:
            sam_file:
        """
        if sam_file_path:
            aln_file = pysam.AlignmentFile(sam_file_path)
            for reference_sequence in aln_file.header['SQ']:
                self.alignment_file_to_reference_sequences_dict[sam_file_path].append(reference_sequence)
            aln_file.close()
        for sample in self.sample_to_vcf_file_dict:            
            for sam_file_path in self.alignment_file_to_reference_sequences_dict:
                if sam_file_path in self.sample_to_alignment_files[str(sample)]:
                    aln_file = pysam.AlignmentFile(sam_file_path)
                    for PG_tag in aln_file.header['PG']:
                        PG_tag['ID'] = str(list(self.RG_ID_dict[sam_file_path].keys())[0])
                        # if PG_tag['ID'] in [PG_tag['ID'] for PG_tag in self.sample_to_PG_header[sample]]
                        RG_tag = {
                                'ID':str(list(self.RG_ID_dict[sam_file_path].keys())[0]), 'SM':str(
                                        self.RG_ID_dict[sam_file_path][
                                            str(list(self.RG_ID_dict[sam_file_path].keys())[0])]['SM']
                                        ),
                                'DS':str(
                                        self.RG_ID_dict[sam_file_path][
                                            str(list(self.RG_ID_dict[sam_file_path].keys())[0])]['DS']
                                        )
                                }
                        # {str(list(self.RG_ID_dict[sam_file_path].keys())[0]): {'DS': 'small7.fastq', 'SM': 'HG001',
                        # 'RG_tags': True, 'outputID': '1'}} self.RG_ID_dict[sam_file_path] self.RG_ID_dict[
                        # sam_file_path][str(list(self.RG_ID_dict[sam_file_path].keys())[0])]['SM'] if not PG_tag[
                        # 'ID'] in [PG_tag['ID'] for PG_tag in self.sample_to_PG_header[sample]]:
                        # self.sample_to_PG_header[sample].append(PG_tag)
                        new_PG_tag: bool = True
                        for PG_header in self.sample_to_PG_header[sample]:
                            if PG_tag['PN'] == PG_header['PN'] and PG_tag['ID'] == PG_header['ID']:
                                new_PG_tag = False
                        if new_PG_tag:
                            self.sample_to_PG_header[sample].append(PG_tag)
                        # self.sample_to_PG_header[sample].append(PG_tag)
                        if not RG_tag['ID'] in [RG_tag['ID'] for RG_tag in self.sample_to_RG_header[sample]]:
                            self.sample_to_RG_header[sample].append(RG_tag)
                    aln_file.close()

    
    def _parse_long_reads_input(self, long_read_input: object, output_directory: str) -> object:
        """

        Args:
            long_read_input (object):
            output_directory:

        Returns:
            object:

        """
        if long_read_input.startswith('http'):
            sys.stderr.write('The long read input is a url. Downloading now.\n')
            long_reads_alignment_path = self._download_file(long_read_input, self.output_directory + '/input_reads')
        else:
            long_reads_alignment_path = long_read_input
        
        combined_long_read_fastq_path = [] # Why is this an array? It gets assigned a string value by os.path.abspath.
        sorted_bam_file_paths = []
        
        if not os.path.exists(long_reads_alignment_path):
            sys.stderr.write("ERROR: Could not find %s.\nUse -i to specify the path of a file containing reads for phasing OR use -i to specify the path of a directory containing the long read files and all files will be processed individually.\n(EX: -i path/minion_run3_GM12878/minion_run3_GM12878_0.fastq OR -i path/minion_run3_GM12878/minion_run3_GM12878_0.sam OR -i path/minion_run3_GM12878/minion_run3_GM12878_0_sorted.bam OR -i path/minion_run3_GM12878/\n\n)." % long_reads_alignment_path)
            raise OSError("File not found")
        
        elif os.path.isdir(long_reads_alignment_path):
            sys.stderr.write('Directory was given as input for read files. Processing all files in %s.\n' % long_reads_alignment_path)
            for aln_file in os.listdir(long_reads_alignment_path):
                qualified_aln_file = '%s/%s' % (os.path.abspath(long_reads_alignment_path), aln_file)
                if os.path.isfile(qualified_aln_file):
                    if str(qualified_aln_file).lower().endswith('.fastq') or str(qualified_aln_file).lower().endswith('.fastq.gz'):
                        sys.stderr.write('%s is a valid fastq file.\n' % qualified_aln_file)
                        if not combined_long_read_fastq_path:
                            combined_long_read_fastq_path = os.path.abspath(
                                    '%s/%s%s' % (
                                            output_directory,
                                            os.path.splitext(os.path.basename(long_reads_alignment_path))[0],
                                            '_combined_fastq.gz')
                                    )
                        with open(combined_long_read_fastq_path, 'a') as combined_fastqs:
                            with open(qualified_aln_file, 'r') as fastqfile:
                                combined_fastqs.write(fastqfile.read())
                    
                    elif str(qualified_aln_file).lower().endswith('.sam') or str(file).lower().endswith('.bam'):
                        sorted_bam_file_paths.append(_prepare_alignment(output_directory, file))
        
        elif os.path.isfile(long_reads_alignment_path):
            if str(long_reads_alignment_path).lower().endswith('.fastq') or str(
                    long_reads_alignment_path
                    ).lower().endswith('.fastq.gz'):
                sys.stderr.write('%s is a valid fastq file.\n' % long_reads_alignment_path)
                combined_long_read_fastq_path = os.path.abspath(long_reads_alignment_path)
            
            elif str(long_reads_alignment_path).lower().endswith('.sam') or str(
                    long_reads_alignment_path
                    ).lower().endswith('.bam'):
                sorted_bam_file_paths.append(_prepare_alignment(output_directory, long_reads_alignment_path))
        else:
            sys.stderr.write('Error: Reads should be in .fastq, .fastq.gz, .sam, or .bam format. %s does not have a correct suffix to be a valid format and is not a directory. Use -i to specify the path of a file containing reads for phasing OR use -i to specify the path of a directory containing the long read files and all files will be processed individually. (EX: -i path/minion_run3_GM12878/minion_run3_GM12878_0.fastq OR -i path/minion_run3_GM12878/minion_run3_GM12878_0.sam OR -i path/minion_run3_GM12878/minion_run3_GM12878_0_sorted.bam OR -i path/minion_run3_GM12878/).' % long_reads_alignment_path)
            raise Exception('Unsupported long read file format.')
        
        return sorted_bam_file_paths, combined_long_read_fastq_path
    
    def _prepare_vcf_file(self, output_directory: object, vcf_file_input: str) -> object:
        """
        Checks the variant file for proper GT headers, bgzip format, and tabix indexing.
        Attempts to sort/bgzip/index if needed.
        Returns path to bgzipped/indexed vcf.
        """
        start_process_time = time.time()
        sys.stderr.write('\n############## Preparing vcf file ##############\n')
        
        if vcf_file_input.startswith('http'):
            sys.stderr.write('The vcf file input is a url. Downloading now.\n')
            vcf_file_input = self._download_file(vcf_file_input, self.output_directory + '/haplotype_information')
        
        if not os.path.isfile(vcf_file_input):
            sys.stderr.write('Could not find %s. Use -v to specify the path of the vcf file to be used as haplotype information for phasing. (EX: -v path/to/GM12878.vcf.gz or --vcf GM12878.vcf).' % vcf_file_input)
            return

        # Check input file for proper format (GT header present, bgzipped, tabix indexed).
        # Correct the file format and index if needed.
        vcf_file_pysam: VariantFile = pysam.VariantFile(vcf_file_input)
        if vcf_file_pysam.format == 'VCF': # What if it's not???
             # Check to be sure the GT header is present.
            if vcf_file_pysam.header.formats.keys().count('GT') == 0:
                sys.stderr.write('The VCF file provided does not have the GT subfield. VCF files must have the GT subfield for all samples in order to extract genotype information. Phased variants should have | in the GT subfield instead of / \n')
                return
            
            # Check for bgzip compression.
            if vcf_file_pysam.compression == 'BGZF':
                vcf_file_path = os.path.abspath(vcf_file_input)
                sys.stderr.write('%s is a valid vcf file.\n' % vcf_file_path)
            # Compress with bgzip if not already done.
            else:
                sys.stderr.write('%s is a valid .vcf file but it is not in bgzip (.vcf.gz) format. VCF files must be compressed with bgzip and indexed with tabix. LRphase will now attempt to run bgzip on %s.\n' % (vcf_file_input, vcf_file_input))
                # Assuming there should be a call to _sort_vcf_file here??
                vcf_file_path = _sort_vcf_file(vcf_file_path)

            # Check for a tabix index and create one if not found.
            vcf_file_index_path = vcf_file_path + '.tbi'    
            if os.path.isfile(vcf_file_index_path):
                sys.stderr.write('Found %s as an index for vcf file' % vcf_file_index_path)
            else:
                sys.stderr.write('%s is a valid .vcf file in bgzip (.vcf.gz) format but an index was not found. Indexing with tabix now.\n' % vcf_file_input)
                pysam.tabix_index(vcf_file_path, preset = 'vcf', force = True)
            # The VariatnFile object is not stored for later use, so close it to free up resources.
            vcf_file_pysam.close()
        else:
            sys.stderr.write("ERROR: %s is not a valid VCF file.\n" % vcf_file_input)
            return

        # Calculate and report time elapsed.
        total_process_time = time.time() - start_process_time       
        sys.stderr.write('Prepared vcf file in %.2f seconds\n' % total_process_time)

        # Return the path to the bgzipped and indexed vcf.
        return vcf_file_path

    
    def _compile_read_groups(
            self, alignment_file_path, sample, ID, sample_description, RG_ID_dict, unique_RG_IDs, ignore_samples
    ):
        if ignore_samples:
            RG_ID_dict[str(alignment_file_path)] = 'ignore_samples'
        else:
            with pysam.AlignmentFile(alignment_file_path, 'rb') as bam_file:
                RG_tags = bam_file.header.get('RG')
            if sample is not None:
                RG_ID_dict[str(alignment_file_path)] = {}
                if ID is not None:
                    RG_ID_dict[str(alignment_file_path)][str(ID)] = {}
                    RG_ID_dict[str(alignment_file_path)][str(ID)]['SM'] = str(sample)
                    RG_ID_dict[str(alignment_file_path)][str(ID)]['outputID'] = str(ID)
                    RG_ID_dict[str(alignment_file_path)][str(ID)]['RG_tags'] = False
                    if sample_description is not None:
                        RG_ID_dict[str(alignment_file_path)][str(ID)]['DS'] = str(sample_description)
                    else:
                        RG_ID_dict[str(alignment_file_path)][str(ID)]['DS'] = str(
                            'LRphase_input_file_' + str(alignment_file_path)
                        )
                else:
                    ID = '0' + str(random.randint(1, 10000))
                    RG_ID_dict[str(alignment_file_path)][str(ID)] = {}
                    RG_ID_dict[str(alignment_file_path)][str(ID)]['SM'] = str(sample)
                    RG_ID_dict[str(alignment_file_path)][str(ID)]['outputID'] = str(ID)
                    RG_ID_dict[str(alignment_file_path)][str(ID)]['RG_tags'] = False
                    if sample_description is not None:
                        RG_ID_dict[str(alignment_file_path)][str(ID)]['DS'] = str(sample_description)
                    else:
                        RG_ID_dict[str(alignment_file_path)][str(ID)]['DS'] = str(
                            'LRphase_input_file_' + str(alignment_file_path)
                        )
                if str(ID) not in unique_RG_IDs:
                    unique_RG_IDs[str(ID)] = {}
                    unique_RG_IDs[str(ID)]['DS'] = str(sample_description)
                    unique_RG_IDs[str(ID)]['SM'] = str(sample)
            elif RG_tags is not None:
                RG_ID_dict, unique_RG_IDs = _unique_RG_IDs_from_RG_tags(
                    RG_ID_dict, unique_RG_IDs, alignment_file_path
                )
            else:
                sys.stderr.write('%s has No RG tags and was not input with sample information. Reads in this file will not be processed. Either re-input this read file with sample information or resubmit with ignore_samples option.\n' % alignment_file_path)
        return RG_ID_dict, unique_RG_IDs


    def _tabulate_sim_results(self):
        """
        Tabulate log-likelihoods and associated error rates for simulation results.

        """
        for phasable_sample in self:
            if phasable_sample.sample == 'HG001':
                for alignment in phasable_sample:
                    phased_read = PhasedRead(alignment, vcf_file = phasable_sample.vcf_file_path, sample = 'HG001', evaluate_true_alignment = True)
                    if phased_read.is_phased_correctly == False:
                        #print(str(phased_read))	
                        #print(str(phased_read.matches_true_alignment))
                        #print("{}\t{}\t{}\n".format(str(phased_read.phase), str(phased_read.is_phased_correctly), str(phased_read.log_likelihood_ratio)))
                        lr = str(phased_read.log_likelihood_ratio)                        
                        if lr in self.lr_err_rate_dict.keys():
                            self.lr_err_rate_dict[lr] += 1
                        else:
                            self.lr_err_rate_dict[lr] = 1
        return

        

def _sort_vcf_file(vcf_file_input: object, vcf_file_output: object = None) -> object:
    """

    Args:
        vcf_file_input: Location of VCF file (absolute path)
        vcf_file_output: Location/name of VCF.bgz output (absolute path)

    Returns path to sorted/bgzipped output file.
    """
    if vcf_file_output == None:
        vcf_file_output = vcf_file_input + '.bgz'
    
    try:
        subprocess.run(
            ["grep -v ^'#' %s | sort -k1,1 -k2,2n | bgzip > %s" % (vcf_file_input, vcf_file_output)],
            check = True, shell = True
        )
        return vcf_file_output
    except Exception as e:
        sys.stderr.write("Error occurred when bgzip was run. Bgzip must be installed on PATH for LRphase to continue. Please check bgzip installation or provide a bgzip compressed vcf file using the -v option (EX: -v path/to/GM12878.vcf.gz). Error: %s\n" % e)
        return
